#############################################################################################
#                                                                                           #
# Author:   GeoPy Team                                                                      #
# Email:    geopy.info@gmail.com                                                            #
# Date:     March 2018                                                                      #
#                                                                                           #
#############################################################################################

# seismic data processing functions

from PyQt5 import QtCore
import sys, os
import numpy as np
#
sys.path.append(os.path.dirname(__file__)[:-4][:-4][:-13])
from cognitivegeo.src.vis.messager import messager as vis_msg


__all__ = ['analysis']

# Seismic data in 2D array (Seis2Dmat) is defined as a 2D array, with
#   Inline ----- First column
#   Xline ----- Second column
#   Z ----- Third column
#   Value ----- Fourth column

# Seismic data in 3D array (Seis3DMat) is defined as a 3D array, with
#   Z ----- First dimension
#   Xline ----- Second dimension
#   Inline ----- Third dimension

# Seismic information (SeisInfo) is defined as a dictionary, with the follow keys,
#   ILStart ----- first inline No.
#   ILEnd ----- last inline No.
#   ILStep ----- step between two adjacent inlines
#   ILNum ----- number of inlines
#   ILRange ----- array of all inlines
#   XLStart ----- first crossline No.
#   XLEnd ----- last crossline No.
#   XLStep ----- step between two adjacent crosslines
#   XLNum ----- number of crosslines
#   XLRange ----- array of all crosslines
#   ZStart ----- top z slice
#   ZEnd ----- bottom z slice
#   ZStep ----- step between two adjacent z slices
#   ZNum ----- number of z slices
#   ZRange ----- array of all z slices


def getSeisInfoFrom2DMat(seis2dmat, inlcol=0, xlcol=1, zcol=2):
    """
    Get basic seismic information from 2D seismic matrix

    Args:
        seis2dmat:  2D mat of seismic data containing at least three columns [IL, XL, Z]
        inlcol:     Index of inline column. Default is the first column (0)
        xlcol:      Index of crossline column. Default is the second column (1)
        zcol:       Index of z column. Default is the third column (2)

    Returns:
        seisinfo:   Seismic information as a dictionary as defined above

    Note:
        Negative z is used in the vertical direction
    """

    # Initialize
    seisinfo = makeSeisInfo()

    # Check size of input 2D matrix
    if np.ndim(seis2dmat) != 2:
        vis_msg.print('ERROR in getSeisInfoFrom2DMat: 2D seismic matrix expected', type='error')
        sys.exit()
    if inlcol < 0 or len(seis2dmat[0,:]) < inlcol:
        vis_msg.print('ERROR in getSeisInfoFrom2DMat: No inline column found in 2D seismic matrix', type='error')
        sys.exit()
    if xlcol < 0 or len(seis2dmat[0,:]) < xlcol:
        vis_msg.print('ERROR in getSeisInfoFrom2DMat: No crossline column found in 2D seismic matrix', type='error')
        sys.exit()
    if zcol < 0 or len(seis2dmat[0,:]) < zcol:
        vis_msg.print('ERROR in getSeisInfoFrom2DMat: No z column found in 2D seismic matrix', type='error')
        sys.exit()

    # Starting from the Z column
    zrange = seis2dmat[:, zcol]
    zstart = zrange[0]
    zend = zrange[-1]
    zstep = -1
    znum = 1
    if zstart != zend:
        zstep = zrange[1] - zrange[0]
        znum = (zend-zstart) / zstep + 1
    zstart = np.round(zstart).astype(np.int32)
    zstep = np.round(zstep).astype(np.int32)
    if zstep == 0:
        zstep = -1
    znum = np.round(znum).astype(np.int32)
    zend = (zstart + (znum-1)*zstep).astype(np.int32)
    zrange = np.linspace(zstart, zend, znum).astype(np.int32)
    # Add to seisinfo
    seisinfo['ZStart'] = zstart
    seisinfo['ZEnd'] = zend
    seisinfo['ZStep'] = zstep
    seisinfo['ZNum'] = znum
    seisinfo['ZRange'] = zrange

    # Then the Crossline column
    xlrange = seis2dmat[:, xlcol]
    xlstart = xlrange[0]
    xlend = xlrange[-1]
    xlstep = 1
    xlnum = 1
    if xlstart != xlend:
        xlstep = xlrange[int(znum)] - xlrange[0]
        xlnum = (xlend-xlstart) / xlstep + 1
    xlstart = np.round(xlstart).astype(np.int32)
    xlstep = np.round(xlstep).astype(np.int32)
    if xlstep == 0:
        xlstep = 1
    xlnum = np.round(xlnum).astype(np.int32)
    xlend = (xlstart + (xlnum-1)*xlstep).astype(np.int32)
    xlrange = np.linspace(xlstart, xlend, xlnum).astype(np.int32)
    # Add to seisinfo
    seisinfo['XLStart'] = xlstart
    seisinfo['XLEnd'] = xlend
    seisinfo['XLStep'] = xlstep
    seisinfo['XLNum'] = xlnum
    seisinfo['XLRange'] = xlrange

    # Finally the Inline column
    inlrange = seis2dmat[:, inlcol]
    inlstart = inlrange[0]
    inlend = inlrange[-1]
    inlstep = 1
    inlnum = 1
    if inlstart != inlend:
        inlstep = inlrange[int(znum*xlnum)] - inlrange[0]
        inlnum = (inlend-inlstart) / inlstep + 1
    inlstart = np.round(inlstart).astype(np.int32)
    inlstep = np.round(inlstep).astype(np.int32)
    inlnum = np.round(inlnum).astype(np.int32)
    if inlstep == 0:
        inlstep = 1
    inlend = (inlstart + (inlnum - 1) * inlstep).astype(np.int32)
    inlrange = np.linspace(inlstart, inlend, inlnum).astype(np.int32)
    # Add to seisinfo
    seisinfo['ILStart'] = inlstart
    seisinfo['ILEnd'] = inlend
    seisinfo['ILStep'] = inlstep
    seisinfo['ILNum'] = inlnum
    seisinfo['ILRange'] = inlrange

    return seisinfo


def checkSeisInfo(seisinfo):
    """
    Check if a seismic info dictionary good

    Args:
        seisinfo:   seismic inforamtion dictionary as defined above

    Return:
        True or false

    Note:
        Negative z is used in the vertical direction
    """

    info = ['ILStart', 'ILEnd', 'ILStep', 'ILNum', 'ILRange',
            'XLStart', 'XLEnd', 'XLStep', 'XLNum', 'XLRange',
            'ZStart', 'ZEnd', 'ZStep', 'ZNum', 'ZRange']

    if seisinfo is None:
        return False

    for f in info:
        if f not in seisinfo.keys():
            #print("Error in checkSeisInfo: %s not found in seisinfo" %f)
            return False

    return True


def makeSeisInfo(inlstart=0, inlstep=1, inlnum=1,
                 xlstart=0, xlstep=1, xlnum=1,
                 zstart=0, zstep=-1, znum=1):
    """
    Make a new seismic information dictionary

    Args:
        inlstart:   first inline. Default is 0
        inlstep:    inline step. Default is 1
        inlnum:     inline number. Default is 1
        xlstart:    first xline. Default is 0
        xlstep:     xline step. Default is 1
        xlnum:      xline number. Default is 1
        zstart:     first z. Default is 0
        zstep:      z step. Default is -1
        znum:       z number. Default is 1

    Return:
         Seismic inforamtion dictionary as defined above
    """

    seisinfo = {}
    seisinfo['ILStart'] = inlstart
    seisinfo['ILEnd'] = inlstart + (inlnum-1)*inlstep
    seisinfo['ILStep'] = inlstep
    seisinfo['ILNum'] = inlnum
    seisinfo['ILRange'] = np.linspace(0, inlnum-1, 1) * inlstep + inlstart
    seisinfo['XLStart'] = xlstart
    seisinfo['XLEnd'] = xlstart + (xlnum-1)*xlstep
    seisinfo['XLStep'] = xlstep
    seisinfo['XLNum'] = xlnum
    seisinfo['XLRange'] = np.linspace(0, inlnum-1, 1) * xlstep + xlstart
    seisinfo['ZStart'] = zstart
    seisinfo['ZEnd'] = zstart + (znum-1) * zstep
    seisinfo['ZStep'] = zstep
    seisinfo['ZNum'] = znum
    seisinfo['ZRange'] = np.linspace(0, inlnum-1, 1) * zstep + zstart

    return seisinfo


def isSeis2DMatConsistentWithSeisInfo(seis2dmat, seisinfo):
    """
    Check if a Seis2DMat is consistent with the given SeisInfo

    Args:
        seis2dmat:  seismic data in 2D array as defined above
        seisinfo:   seismic information as a dictionary as defined above

    Return:
         True or false
    """

    if checkSeisInfo(seisinfo) is False:
        return False
    if np.ndim(seis2dmat) < 2:
        return False
    if np.shape(seis2dmat)[0] != \
            np.int64(seisinfo['ILNum']) * np.int64(seisinfo['XLNum']) * np.int64(seisinfo['ZNum']):
        return False
    return True


def isSeis3DMatConsistentWithSeisInfo(seis3dmat, seisinfo):
    """
    Check if a Seis3DMat is consistent with the given SeisInfo

    Args:
        seis3dmat:  seismic data in 3D array as defined above
        seisinfo:   seismic information as a dictionary as defined above

    Return:
        True or false
    """

    if checkSeisInfo(seisinfo) is False:
        return False
    if np.ndim(seis3dmat) < 3:
        return False
    if np.shape(seis3dmat)[0] != seisinfo['ZNum'] \
            or np.shape(seis3dmat)[1] != seisinfo['XLNum'] \
            or np.shape(seis3dmat)[2] != seisinfo['ILNum']:
        return False
    return True



def createSeisInfoFrom3DMat(seis3dmat, inlstart=0, inlstep=1,
                            xlstart=0, xlstep=1,
                            zstart=0, zstep=-1):
    """
    Create SeisInfo from Seis3DMat

    Args:
        seis3dmat:  3D matrix of the seismic data [Z/XL/IL] as defined above
        inlstart:   first inline No.. Default is 0
        inlstep:    step between two adjacent inlines. Default is 1
        xlstart:    first crosslien No.. Default is 0
        xlstep:     step between two adjacent crosslines. Default is 1
        zstart:     top z slice. Default is 0
        zstep:      step between two adjacent z slices. Default is -1

    Returns:
         seisinfo:  seismic information dictionary as defined above

    Note:
        Negative z is used in the vertical direction
    """

    # Initialize
    seisinfo = makeSeisInfo()

    # Check the dimension of input
    if np.ndim(seis3dmat) != 3:
        vis_msg.print('ERROR in createSeisInfoFrom3DMat: 3D seismic matrix expected', type='error')
        sys.exit()

    znum, xlnum, inlnum = np.shape(seis3dmat)

    inlstart = np.round(inlstart).astype(np.int32)
    inlstep = np.round(inlstep).astype(np.int32)
    inlnum = np.round(inlnum).astype(np.int32)
    if inlstep == 0:
        inlstep = 1
    inlend = (inlstart + inlstep * (inlnum-1)).astype(np.int32)
    inlrange = np.linspace(inlstart, inlend, inlnum).astype(np.int32)
    seisinfo['ILStart'] = inlstart
    seisinfo['ILEnd'] = inlend
    seisinfo['ILStep'] = inlstep
    seisinfo['ILNum'] = inlnum
    seisinfo['ILRange'] = inlrange

    xlstart = np.round(xlstart).astype(np.int32)
    xlstep = np.round(xlstep).astype(np.int32)
    xlnum = np.round(xlnum).astype(np.int32)
    if xlstep == 0:
        xlstep = 1
    xlend = (xlstart + xlstep * (xlnum-1)).astype(np.int32)
    xlrange = np.linspace(xlstart, xlend, xlnum).astype(np.int32)
    seisinfo['XLStart'] = xlstart
    seisinfo['XLEnd'] = xlend
    seisinfo['XLStep'] = xlstep
    seisinfo['XLNum'] = xlnum
    seisinfo['XLRange'] = xlrange

    zstart = np.round(zstart).astype(np.int32)
    zstep = np.round(zstep).astype(np.int32)
    znum = np.round(znum).astype(np.int32)
    if zstep == 0:
        zstep = -1
    zend = (zstart + zstep * (znum - 1)).astype(np.int32)
    zrange = np.linspace(zstart, zend, znum).astype(np.int32)
    seisinfo['ZStart'] = zstart
    seisinfo['ZEnd'] = zend
    seisinfo['ZStep'] = zstep
    seisinfo['ZNum'] = znum
    seisinfo['ZRange'] = zrange

    return seisinfo


def convertSeis2DMatTo3DMat(seis2dmat, datacol=3,
                            inlcol=0, xlcol=1, zcol=2):
    """
    Convert seismic data from 2D array (Seis2DMat) to 3D array (Seis3DMat)

    Argus:
        seis2dmat:  2D matrix of seismic data containing at least four columns
                    [IL, XL, Z, Value ...]
        datacol:    index of data column for conversion (indexing from 0)
                    Default is the fourth column (3)
        inlcol:     index of inline column. Default is the first column (0)
        xlcol:      index of crossline column. Default is the second column (1)
        zcol:       index of z column. Default is the third column (2)

    Returns:
        seis3dmat:  3D seismic matrix [Z/XL/IL] as defined above
    """

    # Check size of input 2D matrix
    if np.ndim(seis2dmat) != 2:
        vis_msg.print('ERROR in convertSeis2DMatTo3DMat: 2D seismic matrix expected', type='error')
        sys.exit()
    if datacol < 0 or len(seis2dmat[0, :]) <= datacol:
        vis_msg.print('ERROR in convertSeis2DMatTo3DMat: No data column found in 2D seismic matrix', type='error')
        sys.exit()
    if inlcol < 0 or len(seis2dmat[0, :]) <= inlcol:
        vis_msg.print('ERROR in convertSeis2DMatTo3DMat: No inline column found in 2D seismic matrix', type='error')
        sys.exit()
    if xlcol < 0 or len(seis2dmat[0, :]) <= xlcol:
        vis_msg.print('ERROR in convertSeis2DMatTo3DMat: No crossline column found in 2D seismic matrix', type='error')
        sys.exit()
    if zcol < 0 or len(seis2dmat[0, :]) <= zcol:
        vis_msg.print('ERROR in convertSeis2DMatTo3DMat: No z column found in 2D seismic matrix', type='error')
        sys.exit()

    # Get the basic information of the seismic survey
    seisinfo = getSeisInfoFrom2DMat(seis2dmat,
                                    inlcol=inlcol, xlcol=xlcol, zcol=zcol)

    inlnum = seisinfo['ILNum']
    xlnum = seisinfo['XLNum']
    znum = seisinfo['ZNum']

    seisdata = seis2dmat[:, datacol]

    seisdata = seisdata.transpose()
    seisdata = np.reshape(seisdata, [inlnum, xlnum, znum])
    seisdata = seisdata.transpose()

    return seisdata


def convertSeis3DMatTo2DMat(seis3dmat, seisinfo=None):
    """
    Convert seismic data in 3D matrix (Seis3DMat) to 2D matrix (Seis2DMat)

    Args:
        seis3dmat:  3D matrix of seismic data [Z/XL/IL]
        seisinfo:   seismic inforamtion dictionary as defined above
                    Auto-generated from 3D seismic matrix if not provided

    Return:
        seis2dmat:  2D seismic matrix of the datacontaining four columns [IL, XL, Z, Value] as defined above
    """

    # Check the dimension of input matrix
    if np.ndim(seis3dmat) != 3:
        vis_msg.print('ERROR in convertSeis3DMatTo2DMat: 3D seismic matrix expected', type='error')
        sys.exit()

    if seisinfo is None:
        vis_msg.print('WARNING in convertSeis3DMatTo2DMat: Survey info auto-generated from 3D seismic matrix',
                      type='warning')
        seisinfo = createSeisInfoFrom3DMat(seis3dmat)

    inlnum = seisinfo['ILNum']
    xlnum = seisinfo['XLNum']
    znum = seisinfo['ZNum']

    if (znum, xlnum, inlnum) != np.shape(seis3dmat):
        vis_msg.print('ERROR in convertSeis3DMatTo2DMat: Dimension not match between 3D matrix and survey info',
                      type='error')
        sys.exit()

    tracenum = (inlnum * xlnum).astype(np.int64)
    samplenum = (tracenum * znum).astype(np.int64)

    seisdata = seis3dmat.transpose()
    seisdata = np.reshape(seisdata, [1, samplenum])
    seisdata = seisdata.transpose()

    # Add inline, crossline, and z information
    inlxlz = convertSeisInfoTo2DMat(seisinfo)

    seis2dmat = np.concatenate((inlxlz, seisdata), axis=1)

    return seis2dmat


def convertSeisInfoTo2DMat(seisinfo):
    """
    Convert seismic info to 2D matrix
    Args:
        seisinfo:   seismic inforamtion dictionary as defined above

    Return:
        seis2dmat:  2D seismic matrix containing three columns [IL, XL, Z]
    """

    if checkSeisInfo(seisinfo) is False:
        vis_msg.print("Error in convertSeisInfoTo2DMat: Seisinfo not correct", type='error')
        sys.exit()

    inlnum = seisinfo['ILNum']
    xlnum = seisinfo['XLNum']
    znum = seisinfo['ZNum']

    tracenum = (inlnum * xlnum).astype(np.int64)
    samplenum = (tracenum * znum).astype(np.int64)

    # Add inline, crossline, and z information
    inlstart = seisinfo['ILStart']
    inlend = seisinfo['ILEnd']
    xlstart = seisinfo['XLStart']
    xlend = seisinfo['XLEnd']
    zstart = seisinfo['ZStart']
    zend = seisinfo['ZEnd']
    #
    all_sample = np.arange(0, samplenum, 1)
    #
    inl = np.linspace(inlstart, inlend, inlnum)
    inl = np.reshape(inl[(all_sample / xlnum / znum).astype(int)], [-1, 1])
    #
    xl = np.linspace(xlstart, xlend, xlnum)
    xl = np.reshape(xl[((all_sample / znum) % xlnum).astype(int)], [-1, 1])
    #
    z = np.linspace(zstart, zend, znum)
    z = np.reshape(z[(all_sample % znum).astype(int)], [-1, 1])

    seis2dmat = np.concatenate((inl, xl, z), axis=1)

    return seis2dmat


def retrieveSeisSampleFrom2DMat(seis2dmat, samples=None, datacol=3,
                                inlcol=0, xlcol=1, zcol=2,
                                verbose=True):
    """
    Retrieve seismic data on specified samples from 2D matrix (Seis2DMat)

    Args:
        seis2dmat:  seismic data in 2D matrix of at least four columns [IL, XL, Z, Value, ...]
        samples:    list of samples in 2D matrix of three columns [IL, XL, Z]
                    All samples are retrieved if not specified
        datacol:    index of data column for retrieval (indexing from 0)
                    Default is the fourth column (3)
        inlcol:     index of inline column. Default is the first column (0)
        xlcol:      index of crossline column. Default is the second column (1)
        zcol:       index of z column. Default is the third column (2)
        verbose:    flag for message display. Default is True

    Return:
        seisdata:   seismic data on specified samples
                    2D matrix of four columns [IL, XL, Z, Value]

    Note:
        Negative z is used in the vertical direction
    """

    # Check input matrix
    if np.ndim(seis2dmat) != 2:
        vis_msg.print('ERROR in retrieveSeisSampleFrom2DMat: 2D seismic matrix expected', type='error')
        sys.exit()
    if datacol < 0 or len(seis2dmat[0,:]) <= datacol:
        vis_msg.print('ERROR in retrieveSeisSampleFrom2DMat: Not data column found in 2D seismic matrix',
                      type='error')
        sys.exit()
    if inlcol < 0 or len(seis2dmat[0,:]) < inlcol:
        vis_msg.print('ERROR in retrieveSeisSampleFrom2DMat: No inline column found in 2D seismic matrix',
                      type='error')
        sys.exit()
    if xlcol < 0 or len(seis2dmat[0,:]) < xlcol:
        vis_msg.print('ERROR in retrieveSeisSampleFrom2DMat: No crossline column found in 2D seismic matrix',
                      type='error')
        sys.exit()
    if zcol < 0 or len(seis2dmat[0,:]) < zcol:
        vis_msg.print('ERROR in retrieveSeisSampleFrom2DMat: No z column found in 2D seismic matrix',
                      type='error')
        sys.exit()

    if samples is None:
        vis_msg.print('WARNING in retrieveSeisSampleFrom2DMat: to retrieve all samples in 2D seismic matrix',
                      type='warning')
        inlloc = seis2dmat[:,inlcol]
        xlloc = seis2dmat[:,xlcol]
        zloc = seis2dmat[:,zcol]
        seisdata = seis2dmat[:, datacol]
        seisdata = np.reshape(seisdata, [len(seisdata), 1])
        seisdata = np.concatenate((inlloc, xlloc, zloc, seisdata), axis=1)
        return seisdata

    if np.ndim(samples) != 2:
        vis_msg.print('ERROR in retrieveSeisSampleFrom2DMat: 2D matrix of samples expected', type='error')
        sys.exit()
    if len(samples[0,:]) < 3:
        vis_msg.print('ERROR in retrieveSeisSampleFrom2DMat: 3 columns expected in 2D sample matrix', type='error')
        sys.exit()

    seisinfo = getSeisInfoFrom2DMat(seis2dmat,
                                    inlcol=inlcol, xlcol=xlcol, zcol=zcol)
    inlstart = seisinfo['ILStart']
    inlstep = seisinfo['ILStep']
    inlnum = seisinfo['ILNum']
    xlstart = seisinfo['XLStart']
    xlstep = seisinfo['XLStep']
    xlnum = seisinfo['XLNum']
    zstart = seisinfo['ZStart']
    zstep = seisinfo['ZStep']
    znum = seisinfo['ZNum']
    if inlnum == 1:
        inlstep = 1
    if xlnum == 1:
        xlstep = 1
    if znum == 1:
        zstep = -1

    nsample, _ = np.shape(samples)
    if verbose:
        print('Retrieve ' + str(nsample) + ' samples')

    seisdata = np.zeros([nsample, 4], np.float32)
    for i in range(nsample):
        #
        if verbose:
            sys.stdout.write('\r>>> Process %.1f%%' % (float(i) / float(nsample) * 100.0))
            sys.stdout.flush()
        #
        inl = samples[i, 0]
        xl = samples[i, 1]
        z = samples[i, 2]

        inlidx = np.round((inl-inlstart)/inlstep).astype(np.int32)
        xlidx = np.round((xl-xlstart)/xlstep).astype(np.int32)
        zidx = np.round((z-zstart)/zstep).astype(np.int32)

        if inlidx>=0 and inlidx<inlnum and xlidx>=0 and xlidx<xlnum \
            and zidx>=0 and zidx<znum :
            idx = inlidx*xlnum*znum+xlidx*znum+zidx
            seisdata[i, 0] = seis2dmat[idx, inlcol]
            seisdata[i, 1] = seis2dmat[idx, xlcol]
            seisdata[i, 2] = seis2dmat[idx, zcol]
            seisdata[i, 3] = seis2dmat[idx, datacol]
    #
    if verbose:
        sys.stdout.write(' Done\n')

    return seisdata


def retrieveSeisSampleFrom3DMat(seis3dmat, samples=None, seisinfo=None,
                                verbose=True):
    """
    Retrieve seismic data on specified samples from 3D matrix (Seis3DMat)

    Args:
        seis3dmat:  seismic data in 3D matrix [Z/XL/IL]
        samples:    list of samples in 2D matrix of three columns [IL, XL, Z]
                    All samples are retrieved if not specified
        seisinfo:   basic information of 3D seismic survey
                    Auto-generated from 3D seismic matrix if not specified
        verbose:    flag for message display. Default is True

    Return:
        seisdata:   seismic data on specified samples
                    2D matrix of four columns [IL, XL, Z, Value]

    Note:
        Negative z is used in the vertical direction
    """

    if np.ndim(seis3dmat) != 3:
        vis_msg.print('ERROR in retrieveSeisSampleFrom3DMat: 3D seismic matrix expected', type='error')
        sys.exit()

    if seisinfo is None:
        vis_msg.print('WARNING in retrieveSeisSampleFrom3DMat: Survey info auto-generated from 3D seismic matrix',
                      type='warning')
        seisinfo = createSeisInfoFrom3DMat(seis3dmat)

    if samples is None:
        vis_msg.print('WARNING in retrieveSeisSampleFrom3DMat: to retrieve all samples in 3D seismic matrix',
                      type='warning')
        return convertSeis3DMatTo2DMat(seis3dmat, seisinfo)

    if np.ndim(samples) != 2:
        vis_msg.print('ERROR in retrieveSeisSampleFrom3DMat: 2D matrix of samples expected', type='error')
        sys.exit()
    if len(samples[0, :]) < 3:
        vis_msg.print('ERROR in retrieveSeisSampleFrom3DMat: 3 columns expected in 2D sample matrix', type='error')
        sys.exit()

    inlstart = seisinfo['ILStart']
    inlstep = seisinfo['ILStep']
    inlnum = seisinfo['ILNum']
    xlstart = seisinfo['XLStart']
    xlstep = seisinfo['XLStep']
    xlnum = seisinfo['XLNum']
    zstart = seisinfo['ZStart']
    zstep = seisinfo['ZStep']
    znum = seisinfo['ZNum']
    if inlnum == 1:
        inlstep = 1
    if xlnum == 1:
        xlstep = 1
    if znum == 1:
        zstep = -1

    nsample, _ = np.shape(samples)
    if verbose:
        print('Retrieve ' + str(nsample) + ' samples')

    seisdata = np.zeros([nsample, 4], np.float32)
    for i in range(nsample):
        #
        if verbose:
            sys.stdout.write('\r>>> Process %.1f%%' % (float(i) / float(nsample) * 100.0))
            sys.stdout.flush()
        #
        inl = samples[i, 0]
        xl = samples[i, 1]
        z = samples[i, 2]

        inlidx = np.round((inl - inlstart) / inlstep).astype(np.int32)
        xlidx = np.round((xl - xlstart) / xlstep).astype(np.int32)
        zidx = np.round((z - zstart) / zstep).astype(np.int32)

        if inlidx >= 0 and inlidx < inlnum and xlidx >= 0 and xlidx < xlnum \
                and zidx >= 0 and zidx < znum:
            idx = inlidx * xlnum * znum + xlidx * znum + zidx
            seisdata[i, 0] = inlidx * inlstep + inlstart
            seisdata[i, 1] = xlidx * xlstep + xlstart
            seisdata[i, 2] = zidx * zstep + zstart
            seisdata[i, 3] = seis3dmat[zidx, xlidx, inlidx]
    #
    if verbose:
        sys.stdout.write(' Done\n')

    return seisdata


def retrieveSeisILSliceFrom2DMat(seis2dmat, inlsls=None, datacol=3,
                                 inlcol=0, xlcol=1, zcol=2,
                                 verbose=True):
    """
    Retrieve seismic data along specified inline slices from 2D matrix (Seis2DMat)

    Args:
        seis2dmat:  seismic data in 2D matrix of at least four columns [IL, XL, Z, Value, ...]
        inlsls:     list of inline slices in array [inl1, inl2, ...]
                    Retrieve all inline slices if not specified
        datacol:    index of data column for retrieval (indexing from 0)
                    Default is the fourth column (3)
        inlcol:     index of inline column. Default is the first column (0)
        xlcol:      index of crossline column. Default is the second column (1)
        zcol:       index of z column. Default is the third column (2)
        verbose:    flag for message display. Default is True

    Return:
        seisdata:   seismic data along specified inline slices
                    2D matrix of four columns [IL, XL, Z, Value]

    Note:
        Negative z is used in the vertical direction
    """

    # Check input matrix
    if np.ndim(seis2dmat) != 2:
        vis_msg.print('ERROR in retrieveSeisILSliceFrom2DMat: 2D seismic matrix expected', type='error')
        sys.exit()
    if datacol < 0 or len(seis2dmat[0, :]) <= datacol:
        vis_msg.print('ERROR in retrieveSeisILSliceFrom2DMat: No data columns found in 2D seismic matrix',
                      type='error')
        sys.exit()
    if inlcol < 0 or len(seis2dmat[0,:]) < inlcol:
        vis_msg.print('ERROR in retrieveSeisILSliceFrom2DMat: No inline column found in 2D seismic matrix',
                      type='error')
        sys.exit()
    if xlcol < 0 or len(seis2dmat[0,:]) < xlcol:
        vis_msg.print('ERROR in retrieveSeisILSliceFrom2DMat: No crossline column found in 2D seismic matrix',
                      type='error')
        sys.exit()
    if zcol < 0 or len(seis2dmat[0,:]) < zcol:
        vis_msg.print('ERROR in retrieveSeisILSliceFrom2DMat: No z column found in 2D seismic matrix', type='error')
        sys.exit()

    if inlsls is None:
        vis_msg.print('WARNING in retrieveSeisILSliceFrom2DMat: to retrieve all inline slices in 2D seismic matrix',
                      type='warning')
        inlloc = seis2dmat[:, inlcol]
        xlloc = seis2dmat[:, xlcol]
        zloc = seis2dmat[:, zcol]
        seisdata = seis2dmat[:, datacol]
        seisdata = np.reshape(seisdata, [len(seisdata), 1])
        seisdata = np.concatenate((inlloc, xlloc, zloc, seisdata), axis=1)
        return seisdata

    if np.ndim(inlsls) != 1:
        vis_msg.print('ERROR in retrieveSeisILSliceFrom2DMat: 1D array of inline slices expected', type='error')
        sys.exit()

    seisinfo = getSeisInfoFrom2DMat(seis2dmat,
                                    inlcol=inlcol, xlcol=xlcol, zcol=zcol)

    inl3dmat = convertSeis2DMatTo3DMat(seis2dmat, datacol=inlcol,
                                       inlcol=inlcol, xlcol=xlcol, zcol=zcol)
    xl3dmat = convertSeis2DMatTo3DMat(seis2dmat, datacol=xlcol,
                                      inlcol=inlcol, xlcol=xlcol, zcol=zcol)
    z3dmat = convertSeis2DMatTo3DMat(seis2dmat, datacol=zcol,
                                     inlcol=inlcol, xlcol=xlcol, zcol=zcol)
    seis3dmat = convertSeis2DMatTo3DMat(seis2dmat, datacol=datacol,
                                        inlcol=inlcol, xlcol=xlcol, zcol=zcol)

    inlstart = seisinfo['ILStart']
    inlstep = seisinfo['ILStep']
    znum, xlnum, inlnum = np.shape(seis3dmat)
    if inlnum == 1:
        inlstep = 1

    ninlsls = len(inlsls)
    if verbose:
        print('Retrieve ' + str(ninlsls) + ' inline slices')

    seisinl = np.zeros([znum, xlnum, ninlsls], np.float32)
    seisxl = np.zeros([znum, xlnum, ninlsls], np.float32)
    seisz = np.zeros([znum, xlnum, ninlsls], np.float32)
    seisdata = np.zeros([znum, xlnum, ninlsls], np.float32)
    for i in range(ninlsls):
        #
        if verbose:
            sys.stdout.write('\r>>> Process %.1f%%' % (float(i) / float(ninlsls) * 100.0))
            sys.stdout.flush()
        #
        inl = inlsls[i]
        idx = np.round((inl-inlstart)/inlstep).astype(np.int32)
        if idx>=0 and idx<inlnum:
            seisinl[:, :, i] = inl3dmat[:, :, idx]
            seisxl[:, :, i] = xl3dmat[:, :, idx]
            seisz[:, :, i] = z3dmat[:, :, idx]
            seisdata[:, :, i] = seis3dmat[:, :, idx]
    #
    if verbose:
        sys.stdout.write(' Done\n')

    seisinl = seisinl.transpose()
    seisinl = np.reshape(seisinl, [1, znum * xlnum * ninlsls])
    seisinl = seisinl.transpose()
    seisxl = seisxl.transpose()
    seisxl = np.reshape(seisxl, [1, znum * xlnum * ninlsls])
    seisxl = seisxl.transpose()
    seisz = seisz.transpose()
    seisz = np.reshape(seisz, [1, znum * xlnum * ninlsls])
    seisz = seisz.transpose()
    seisdata = seisdata.transpose()
    seisdata = np.reshape(seisdata, [1, znum*xlnum*ninlsls])
    seisdata = seisdata.transpose()

    seisdata = np.concatenate((seisinl, seisxl, seisz, seisdata), axis=1)

    return seisdata


def retrieveSeisILSliceFrom3DMat(seis3dmat, inlsls=None, seisinfo=None,
                                 verbose=True):
    """
    Retrieve seismic data along specified inline slices from 3D matrix (Seis3DMat)

    Args:
        seisddmat:  seismic data in 3D matrix [Z/XL/IL]
        inlsls:     list of inline slices in array [inl1, inl2, ...]
                    Retrieve all inline slices if not specified
        seisinfo:   basic information of 3D seismic survey
                    Auto-generated from 3D seismic matrix if not specified
        verbose:    flag for message display. Default is True

    Return:
        seisdata:   seismic data along specified inline slices
                    2D matrix of four columns [IL, XL, Z, Value]

    Note:
        Negative z is used in the vertical direction
    """

    if np.ndim(seis3dmat) != 3:
        vis_msg.print('ERROR in retrieveSeisILSliceFrom3DMat: 3D seismic matrix expected', type='error')
        sys.exit()

    if seisinfo is None:
        vis_msg.print('WARNING in retrieveSeisILSliceFrom3DMat: Survey info auto-generated from 3D seismic matrix',
                      type='warning')
        seisinfo = createSeisInfoFrom3DMat(seis3dmat)


    if inlsls is None:
        vis_msg.print('WARNING in retrieveSeisILSliceFrom3DMat: to retrieve all inline slices in 3D seismic matrix',
                      type='warning')
        seis2dmat = convertSeis3DMatTo2DMat(seis3dmat, seisinfo)
        seisloc = seis2dmat[:, 0:3]
        seisdata = seis2dmat[:, 3]
        seisdata = np.reshape(seisdata, [len(seisdata), 1])
        seisdata = np.concatenate((seisloc, seisdata), axis=1)
        return seisdata

    if np.ndim(inlsls) != 1:
        vis_msg.print('ERROR in retrieveSeisILSliceFrom3DMat: 1D array of inline slices expected', tpye='error')
        sys.exit()

    inlstart = seisinfo['ILStart']
    inlstep = seisinfo['ILStep']
    znum, xlnum, inlnum = np.shape(seis3dmat)
    if inlnum == 1:
        inlstep = 1

    ninlsls = len(inlsls)
    if verbose:
        print('Retrieve ' + str(ninlsls) + ' inline slices')

    seisdata = np.zeros([znum, xlnum, ninlsls], np.float32)
    for i in range(ninlsls):
        #
        if verbose:
            sys.stdout.write('\r>>> Process %.1f%%' % (float(i+1) / float(ninlsls) * 100.0))
            sys.stdout.flush()
        #
        inl = inlsls[i]
        idx = np.round((inl - inlstart) / inlstep).astype(np.int32)
        if idx >= 0 and idx < inlnum:
            seisdata[:, :, i] = seis3dmat[:, :, idx]
    #
    if verbose:
        sys.stdout.write(' Done\n')

    all_sample = np.arange(0, ninlsls*xlnum*znum)
    seisinl = inlsls
    seisinl = np.reshape(seisinl[(all_sample/xlnum/znum).astype(int)], [-1, 1])
    seisxl = seisinfo['XLRange']
    seisxl = np.reshape(seisxl[((all_sample/znum) % xlnum).astype(int)], [-1, 1])
    seisz = seisinfo['ZRange']
    seisz = np.reshape(seisz[(all_sample % znum).astype(int)], [-1, 1])
    #
    seisdata = seisdata.transpose()
    seisdata = np.reshape(seisdata, [1, znum * xlnum * ninlsls])
    seisdata = seisdata.transpose()

    seisdata = np.concatenate((seisinl, seisxl, seisz, seisdata), axis=1)

    return seisdata


def cutSeisILSliceFrom2DMat(seis2dmat, inlsls=None,
                            imageheight=100, imagewidth=100,
                            xlstart=None, xlend=None, xlstep=None,
                            zstart=None, zend=None, zstep=None,
                            datacol=3, inlcol=0, xlcol=1, zcol=2,
                            verbose=True):
    """
    Cut seismic inline slices from Seis2DMat

    Args:
        seis2dmat:      seismic data in 2D array
        inlsls:         inline slices for cutting. Default is None, to cut all slices
        imageheight:    image height. Default is 100
        imagewidth:     image width. Default is 100
        xlstart:        first xline. Default is None, to cut from the first xline
        xlend:          last xline. Default is None, to cut until the last xline
        xlstep:         xline step. Default is None, to cut with the xline step
        zstart:         first z. Default is None, to cut from the z
        zend:           last z. Default is None, to cut until the last z
        zstep:          z step. Default is None, to cut with the z step
        datacol:        data column index in Seis2DMat. Default is 3
        inlcol:         inline column index in Seis2DMat. Default is 0
        xlcol:          xline column index in Seis2DMat. Default is 1
        zcol:           z column index in Seis2DMat. Default is 2
        verbose:        verbose option for progress display. Default is True

    Return:
        A 4D array of the cut inline slices, with
            Dimension 1 ----- inline
            Dimension 2 ----- crossline
            Dimension 3 ----- z
            Dimension 4 ----- pixel
    """

    seisinfo = getSeisInfoFrom2DMat(seis2dmat,
                                    inlcol=inlcol, xlcol=xlcol, zcol=zcol)
    seis3dmat = convertSeis2DMatTo3DMat(seis2dmat, datacol=datacol,
                                        inlcol=inlcol, xlcol=xlcol, zcol=zcol)
    #
    if inlsls is None:
        vis_msg.print('WARNING in cutSeisILSliceFrom2DMat: to cut all inline slices in 2D seismic matrix',
                      type='error')
        inlsls = seisinfo['ILRange']
    if xlstart is None:
        xlstart = seisinfo['XLStart']
    if xlend is None:
        xlend = seisinfo['XLEnd']
    if xlstep is None:
        xlstep = seisinfo['XLStep']
    if zstart is None:
        zstart = seisinfo['ZStart']
    if zend is None:
        zend = seisinfo['ZEnd']
    if zstep is None:
        zstep = seisinfo['ZStep']

    #
    xlidx = np.arange(xlstart, xlend+1, xlstep)
    xlidx = np.round((xlidx - seisinfo['XLStart'])/seisinfo['XLStep']).astype(np.int32)
    xlidx = np.unique(xlidx)
    outidx = []
    for i in range(len(xlidx)):
        if xlidx[i] < 0 or xlidx[i] + imagewidth > seisinfo['XLNum']:
            outidx.append(i)
    xlidx = np.delete(xlidx, outidx)
    #
    zidx = np.arange(zstart, zend, zstep)
    zidx = np.round((zidx - seisinfo['ZStart']) / seisinfo['ZStep']).astype(np.int32)
    zidx = np.unique(zidx)
    outidx = []
    for i in range(len(zidx)):
        if zidx[i] < 0 or zidx[i] + imageheight > seisinfo['ZNum']:
            outidx.append(i)
    zidx = np.delete(zidx, outidx)
    #
    ninlsls = len(inlsls)
    if verbose:
        print('Cut ' + str(ninlsls) + ' inline slices')
    nxlidx = len(xlidx)
    nzidx = len(zidx)
    #
    seisilslicepatch = np.zeros([ninlsls, nxlidx, nzidx, imageheight, imagewidth])
    #
    for i in range(ninlsls):
        #
        if verbose:
            sys.stdout.write('\r>>> Process %.1f%%' % (float(i) / float(ninlsls) * 100.0))
            sys.stdout.flush()
        #
        inl = inlsls[i]
        idx = np.round((inl-seisinfo['ILStart'])/seisinfo['ILStep']).astype(np.int32)
        for j in range(nxlidx):
            for k in range(nzidx):
                seisilslicepatch[i, j, k, :, :] = \
                    seis3dmat[zidx[k]:zidx[k]+imageheight, xlidx[j]:xlidx[j]+imagewidth, idx]
    #
    if verbose:
        sys.stdout.write(' Done\n')

    return np.reshape(seisilslicepatch, [ninlsls, nxlidx, nzidx, imageheight*imagewidth])


def retrieveSeisXLSliceFrom2DMat(seis2dmat, xlsls=None, datacol=3,
                                 inlcol=0, xlcol=1, zcol=2,
                                 verbose=True):
    """
    Retrieve seismic data along specified crossline slices from 2D matrix (Seis2DMat)

    Args:
        seis2dmat:  seismic data in 2D matrix of at least four columns [IL, XL, Z, Value, ...]
        xlsls:      list of crossline slices in array [xl1, xl2, ...]
                    Retrieve all crossline slices if not specified
        datacol:    index of data column for retrieval (indexing from 0)
                    Default is the fourth column (3)
        inlcol:     index of inline column. Default is the first column (0)
        xlcol:      index of crossline column. Default is the second column (1)
        zcol:       index of z column. Default is the third column (2)
        verbose:    flag for message display. Default is True

    Return:
        seisdata:   seismic data along specified crossline slices
                    as 2D matrix of four columns [IL, XL, Z, Value]

    Note:
        Negative z is used in the vertical direction
    """

    # Check input matrix
    if np.ndim(seis2dmat) != 2:
        vis_msg.print('ERROR in retrieveSeisXLSliceFrom2DMat: 2D seismic matrix expected', type='error')
        sys.exit()
    if datacol < 0 or len(seis2dmat[0, :]) <= datacol:
        vis_msg.print('ERROR in retrieveSeisXLSliceFrom2Dmat: Not data column found in 2D seismic matrix',
                      type='error')
        sys.exit()
    if inlcol < 0 or len(seis2dmat[0, :]) <= inlcol:
        vis_msg.print('ERROR in retrieveSeisXLSliceFrom2Dmat: Not inline column found in 2D seismic matrix',
                      type='error')
        sys.exit()
    if xlcol < 0 or len(seis2dmat[0, :]) <= xlcol:
        vis_msg.print('ERROR in retrieveSeisXLSliceFrom2Dmat: Not crossline column found in 2D seismic matrix',
                      type='error')
        sys.exit()
    if zcol < 0 or len(seis2dmat[0, :]) <= zcol:
        vis_msg.print('ERROR in retrieveSeisXLSliceFrom2Dmat: Not z column found in 2D seismic matrix', type='error')
        sys.exit()

    if xlsls is None:
        vis_msg.print('WARNING in retrieveSeisXLSliceFrom2DMat: to retrieve all crossline slices in 2D seismic matrix',
                      type='warning')
        inlloc = seis2dmat[:, inlcol]
        xlloc = seis2dmat[:, xlcol]
        zloc = seis2dmat[:, zcol]
        seisdata = seis2dmat[:, datacol]
        seisdata = np.reshape(seisdata, [len(seisdata), 1])
        seisdata = np.concatenate((inlloc, xlloc, zloc, seisdata), axis=1)
        return seisdata

    if np.ndim(xlsls) != 1:
        vis_msg.print('ERROR in retrieveSeisXLSliceFrom2DMat: 1D array of crossline slices expected', type='error')
        sys.exit()

    seisinfo = getSeisInfoFrom2DMat(seis2dmat,
                                    inlcol=inlcol, xlcol=xlcol, zcol=zcol)

    inl3dmat = convertSeis2DMatTo3DMat(seis2dmat, datacol=inlcol,
                                       inlcol=inlcol, xlcol=xlcol, zcol=zcol)
    xl3dmat = convertSeis2DMatTo3DMat(seis2dmat, datacol=xlcol,
                                      inlcol=inlcol, xlcol=xlcol, zcol=zcol)
    z3dmat = convertSeis2DMatTo3DMat(seis2dmat, datacol=zcol,
                                     inlcol=inlcol, xlcol=xlcol, zcol=zcol)
    seis3dmat = convertSeis2DMatTo3DMat(seis2dmat, datacol=datacol,
                                        inlcol=inlcol, xlcol=xlcol, zcol=zcol)

    xlstart = seisinfo['XLStart']
    xlstep = seisinfo['XLStep']
    znum, xlnum, inlnum = np.shape(seis3dmat)
    if xlnum == 1:
        xlstep = 1

    nxlsls = len(xlsls)
    if verbose:
        print('Retrieve ' + str(nxlsls) + ' crossline slices')

    seisinl = np.zeros([znum, nxlsls, inlnum], np.float32)
    seisxl = np.zeros([znum, nxlsls, inlnum], np.float32)
    seisz = np.zeros([znum, nxlsls, inlnum], np.float32)
    seisdata = np.zeros([znum, nxlsls, inlnum], np.float32)
    for i in range(nxlsls):
        #
        if verbose:
            sys.stdout.write('\r>>> Process %.1f%%' % (float(i) / float(nxlsls) * 100.0))
            sys.stdout.flush()
        #
        xl = xlsls[i]
        idx = np.round((xl - xlstart) / xlstep).astype(np.int32)
        if idx >= 0 and idx < xlnum:
            seisinl[:, i, :] = inl3dmat[:, idx, :]
            seisxl[:, i, :] = xl3dmat[:, idx, :]
            seisz[:, i, :] = z3dmat[:, idx, :]
            seisdata[:, i, :] = seis3dmat[:, idx, :]
    #
    if verbose:
        sys.stdout.write(' Done\n')

    seisinl = seisinl.transpose()
    seisinl = np.reshape(seisinl, [1, znum * nxlsls * inlnum])
    seisinl = seisinl.transpose()
    seisxl = seisxl.transpose()
    seisxl = np.reshape(seisxl, [1, znum * nxlsls * inlnum])
    seisxl = seisxl.transpose()
    seisz = seisz.transpose()
    seisz = np.reshape(seisz, [1, znum * nxlsls * inlnum])
    seisz = seisz.transpose()
    seisdata = seisdata.transpose()
    seisdata = np.reshape(seisdata, [1, znum * nxlsls * inlnum])
    seisdata = seisdata.transpose()

    seisdata = np.concatenate((seisinl, seisxl, seisz, seisdata), axis=1)

    return seisdata


def retrieveSeisXLSliceFrom3DMat(seis3dmat, xlsls=None, seisinfo=None,
                                 verbose=True):
    """
    Retrieve seismic data along specified crossline slices from 3D matrix (Seis3DMat)

    Args:
        seisddmat:  seismic data in 3D matrix [Z/XL/IL]
        xlsls:     list of crossline slices in array [xl1, xl2, ...]
                    Retrieve all crossline slices if not specified
        seisinfo:   basic information of 3D seismic survey
                    Auto-generated from 3D seismic matrix if not specified
        verbose:    flag for message display. Default is True

    Return:
        seisdata:   seismic data along specified inline slices
                    2D matrix of four columns [IL, XL, Z, Value]

    Note:
        Negative z is used in the vertical direction
    """

    if np.ndim(seis3dmat) != 3:
        vis_msg.print('ERROR in retrieveSeisXLSliceFrom3DMat: 3D seismic matrix expected', type='error')
        sys.exit()

    if seisinfo is None:
        vis_msg.print('WARNING in retrieveSeisXLSliceFrom3DMat: Survey info auto-generated from 3D seismic matrix',
                      type='warning')
        seisinfo = createSeisInfoFrom3DMat(seis3dmat)


    if xlsls is None:
        vis_msg.print('WARNING in retrieveSeisXLSliceFrom3DMat: to retrieve all crossline slices in 3D seismic matrix',
                      type='warning')
        seis2dmat = convertSeis3DMatTo2DMat(seis3dmat, seisinfo)
        seisloc = seis2dmat[:, 0:3]
        seisdata = seis2dmat[:, 3]
        seisdata = np.reshape(seisdata, [len(seisdata), 1])
        seisdata = np.concatenate((seisloc, seisdata), axis=1)
        return seisdata

    if np.ndim(xlsls) != 1:
        vis_msg.print('ERROR in retrieveSeisXLSliceFrom3DMat: 1D array of crossline slices expected', type='error')
        sys.exit()

    xlstart = seisinfo['XLStart']
    xlstep = seisinfo['XLStep']
    znum, xlnum, inlnum = np.shape(seis3dmat)
    if xlnum == 1:
        xlstep = 1

    nxlsls = len(xlsls)
    if verbose:
        print('Retrieve ' + str(nxlsls) + ' crossline slices')

    seisdata = np.zeros([znum, nxlsls, inlnum], np.float32)
    for i in range(nxlsls):
        #
        if verbose:
            sys.stdout.write('\r>>> Process %.1f%%' % (float(i+1) / float(nxlsls) * 100.0))
            sys.stdout.flush()
        #
        xl = xlsls[i]
        idx = np.round((xl - xlstart) / xlstep).astype(np.int32)
        if idx >= 0 and idx < xlnum:
            seisdata[:, i, :] = seis3dmat[:, idx, :]
    #
    if verbose:
        sys.stdout.write(' Done\n')

    all_sample = np.arange(0, inlnum * nxlsls * znum)
    seisinl = seisinfo['ILRange']
    seisinl = np.reshape(seisinl[(all_sample / nxlsls / znum).astype(int)], [-1, 1])
    seisxl = xlsls
    seisxl = np.reshape(seisxl[((all_sample / znum) % nxlsls).astype(int)], [-1, 1])
    seisz = seisinfo['ZRange']
    seisz = np.reshape(seisz[(all_sample % znum).astype(int)], [-1, 1])
    #
    seisdata = seisdata.transpose()
    seisdata = np.reshape(seisdata, [1, znum * nxlsls * inlnum])
    seisdata = seisdata.transpose()

    seisdata = np.concatenate((seisinl, seisxl, seisz, seisdata), axis=1)

    return seisdata


def cutSeisXLSliceFrom2DMat(seis2dmat,xlsls=None,
                            imageheight=100, imagewidth=100,
                            inlstart=None, inlend=None, inlstep=None,
                            zstart=None, zend=None, zstep=None,
                            datacol=3, inlcol=0, xlcol=1, zcol=2,
                            verbose=True):
    """
    Cut seismic crossline slices from Seis2DMat

    Args:
        seis2dmat:      seismic data in 2D array
        xlsls:          crossline slices for cutting. Default is None, to cut all slices
        imageheight:    image height. Default is 100
        imagewidth:     image width. Default is 100
        inlstart:       first inline. Default is None, to cut from the first inline
        inlend:         last inline. Default is None, to cut until the last inline
        inlstep:        inline step. Default is None, to cut with the inline step
        zstart:         first z. Default is None, to cut from the z
        zend:           last z. Default is None, to cut until the last z
        zstep:          z step. Default is None, to cut with the z step
        datacol:        data column index in Seis2DMat. Default is 3
        inlcol:         inline column index in Seis2DMat. Default is 0
        xlcol:          xline column index in Seis2DMat. Default is 1
        zcol:           z column index in Seis2DMat. Default is 2
        verbose:        verbose option for progress display. Default is True

    Return:
        A 4D array of the cut inline slices, with
            Dimension 1 ----- inline
            Dimension 2 ----- crossline
            Dimension 3 ----- z
            Dimension 4 ----- pixel
    """


    seisinfo = getSeisInfoFrom2DMat(seis2dmat,
                                    inlcol=inlcol, xlcol=xlcol, zcol=zcol)
    seis3dmat = convertSeis2DMatTo3DMat(seis2dmat, datacol=datacol,
                                        inlcol=inlcol, xlcol=xlcol, zcol=zcol)
    #
    if xlsls is None:
        vis_msg.print('WARNING in cutSeisXLSliceFrom2DMat: to cut all crossline slices in 2D seismic matrix',
                      type='warning')
        inlsls = seisinfo['ILRange']
    if inlstart is None:
        inlstart = seisinfo['ILStart']
    if inlend is None:
        inlend = seisinfo['ILEnd']
    if inlstep is None:
        inlstep = seisinfo['ILStep']
    if zstart is None:
        zstart = seisinfo['ZStart']
    if zend is None:
        zend = seisinfo['ZEnd']
    if zstep is None:
        zstep = seisinfo['ZStep']

    #
    inlidx = np.arange(inlstart, inlend+1, inlstep)
    inlidx = np.round((inlidx - seisinfo['ILStart'])/seisinfo['ILStep']).astype(np.int32)
    inlidx = np.unique(inlidx)
    outidx = []
    for i in range(len(inlidx)):
        if inlidx[i] < 0 or inlidx[i] + imagewidth > seisinfo['ILNum']:
            outidx.append(i)
    inlidx = np.delete(inlidx, outidx)
    #
    zidx = np.arange(zstart, zend, zstep)
    zidx = np.round((zidx - seisinfo['ZStart']) / seisinfo['ZStep']).astype(np.int32)
    zidx = np.unique(zidx)
    outidx = []
    for i in range(len(zidx)):
        if zidx[i] < 0 or zidx[i] + imageheight > seisinfo['ZNum']:
            outidx.append(i)
    zidx = np.delete(zidx, outidx)
    #
    nxlsls = len(xlsls)
    if verbose:
        print('Cut ' + str(nxlsls) + ' crossline slices')
    ninlidx = len(inlidx)
    nzidx = len(zidx)
    #
    seisxlslicepatch = np.zeros([ninlidx, nxlsls, nzidx, imageheight, imagewidth])
    #
    for i in range(nxlsls):
        #
        if verbose:
            sys.stdout.write('\r>>> Process %.1f%%' % (float(i) / float(nxlsls) * 100.0))
            sys.stdout.flush()
        #
        xl = xlsls[i]
        idx = np.round((xl-seisinfo['XLStart'])/seisinfo['ILStep']).astype(np.int32)
        for j in range(ninlidx):
            for k in range(nzidx):
                seisxlslicepatch[j, i, k, :, :] = \
                    seis3dmat[zidx[k]:zidx[k]+imageheight, idx, inlidx[j]:inlidx[j]+imagewidth]
    #
    if verbose:
        sys.stdout.write(' Done\n')

    return np.reshape(seisxlslicepatch, [ninlidx, nxlsls, nzidx, imageheight*imagewidth])


def retrieveSeisZSliceFrom2DMat(seis2dmat, zsls=None, datacol=3,
                                inlcol=0, xlcol=1, zcol=2,
                                verbose=True):
    """
    Retrieve seismic data along specified z slices from 2D matrix (Seis2DMat)

    Args:
        seis2dmat:  seismic data in 2D matrix of at least four columns [IL, XL, Z, Value]
        zsls:       list of z slices in array [z1, z2, ...]
                    Retrieve all z slices if not specified
        datacol:    index of data column for retrieval (indexing from 0)
                    Default is the fourth column (3)
        inlcol:     index of inline column. Default is the first column (0)
        xlcol:      index of crossline column. Default is the second column (1)
        zcol:       index of z column. Default is the third column (2)
        verbose:    flag for message display. Default is True

    Return:
        seisdata:   seismic data along specified z slices
                    2D matrix of four columns [IL, XL, Z, Value]

    Note:
        Negative z is used in the vertical direction
    """

    # Check input matrix
    if np.ndim(seis2dmat) != 2:
        vis_msg.print('ERROR in retrieveSeisZSliceFrom2DMat: 2D seismic matrix expected', type='error')
        sys.exit()
    if datacol < 0 or len(seis2dmat[0, :]) <= datacol:
        vis_msg.print('ERROR in retrieveSeisZSliceFrom2DMat: Not data column found in 2D seismic matrix',
                      type='error')
        sys.exit()
    if inlcol < 0 or len(seis2dmat[0, :]) <= inlcol:
        vis_msg.print('ERROR in retrieveSeisZSliceFrom2DMat: Not inline column found in 2D seismic matrix',
                      type='error')
        sys.exit()
    if xlcol < 0 or len(seis2dmat[0, :]) <= xlcol:
        vis_msg.print('ERROR in retrieveSeisZSliceFrom2DMat: Not crossline column found in 2D seismic matrix',
                      type='error')
        sys.exit()
    if zcol < 0 or len(seis2dmat[0, :]) <= zcol:
        vis_msg.print('ERROR in retrieveSeisZSliceFrom2DMat: Not z column found in 2D seismic matrix', tpye='error')
        sys.exit()

    if zsls is None:
        vis_msg.print('WARNING in retrieveSeisZSliceFrom2DMat: to retrieve all z slices in 2D seismic matrix',
                      type='warning')
        inlloc = seis2dmat[:, inlcol]
        xlloc = seis2dmat[:, xlcol]
        zloc = seis2dmat[:, zcol]
        seisdata = seis2dmat[:, datacol]
        seisdata = np.reshape(seisdata, [len(seisdata), 1])
        seisdata = np.concatenate((inlloc, xlloc, zloc, seisdata), axis=1)
        return seisdata

    if np.ndim(zsls) != 1:
        vis_msg.print('ERROR in retrieveSeisZSliceFrom2DMat: 1D array for z slices expected', type='error')
        sys.exit()

    seisinfo = getSeisInfoFrom2DMat(seis2dmat,
                                    inlcol=inlcol, xlcol=xlcol, zcol=zcol)

    inl3dmat = convertSeis2DMatTo3DMat(seis2dmat, datacol=inlcol,
                                       inlcol=inlcol, xlcol=xlcol, zcol=zcol)
    xl3dmat = convertSeis2DMatTo3DMat(seis2dmat, datacol=xlcol,
                                      inlcol=inlcol, xlcol=xlcol, zcol=zcol)
    z3dmat = convertSeis2DMatTo3DMat(seis2dmat, datacol=2,
                                     inlcol=inlcol, xlcol=xlcol, zcol=zcol)
    seis3dmat = convertSeis2DMatTo3DMat(seis2dmat, datacol=datacol,
                                        inlcol=inlcol, xlcol=xlcol, zcol=zcol)

    zstart = seisinfo['ZStart']
    zstep = seisinfo['ZStep']
    znum, xlnum, inlnum = np.shape(seis3dmat)
    if znum == 1:
        zstep = -1

    nzsls = len(zsls)
    if verbose:
        print('Retrieve ' + str(nzsls) + ' z slices')

    seisinl = np.zeros([nzsls, xlnum, inlnum], np.float32)
    seisxl = np.zeros([nzsls, xlnum, inlnum], np.float32)
    seisz = np.zeros([nzsls, xlnum, inlnum], np.float32)
    seisdata = np.zeros([nzsls, xlnum, inlnum], np.float32)
    for i in range(nzsls):
        #
        if verbose:
            sys.stdout.write('\r>>> Process %.1f%%' % (float(i) / float(nzsls) * 100.0))
            sys.stdout.flush()
        #
        z = zsls[i]
        idx = np.round((z - zstart) / zstep).astype(np.int32)
        if idx >= 0 and idx < znum:
            seisinl[i, :, :] = inl3dmat[idx, :, :]
            seisxl[i, :, :] = xl3dmat[idx, :, :]
            seisz[i, :, :] = z3dmat[idx, :, :]
            seisdata[i, :, :] = seis3dmat[idx, :, :]
    #
    if verbose:
        sys.stdout.write(' Done\n')

    seisinl = seisinl.transpose()
    seisinl = np.reshape(seisinl, [1, nzsls * xlnum * inlnum])
    seisinl = seisinl.transpose()
    seisxl = seisxl.transpose()
    seisxl = np.reshape(seisxl, [1, nzsls * xlnum * inlnum])
    seisxl = seisxl.transpose()
    seisz = seisz.transpose()
    seisz = np.reshape(seisz, [1, nzsls * xlnum * inlnum])
    seisz = seisz.transpose()
    seisdata = seisdata.transpose()
    seisdata = np.reshape(seisdata, [1, nzsls * xlnum * inlnum])
    seisdata = seisdata.transpose()

    seisdata = np.concatenate((seisinl, seisxl, seisz, seisdata), axis=1)

    return seisdata


def retrieveSeisZSliceFrom3DMat(seis3dmat, zsls=None, seisinfo=None,
                                verbose=True):
    """
    Retrieve seismic data along specified z slices from 3D matrix (Seis3DMat)

    Args:
        seisddmat:  seismic data in 3D matrix [Z/XL/IL]
        zlsls:      list of z slices in array [z1, z2, ...]
                    Retrieve all z slices if not specified
        seisinfo:   basic information of 3D seismic survey
                    Auto-generated from 3D seismic matrix if not specified
        verbose:    flag for message display. Default is True

    Return:
        seisdata:   seismic data along specified inline slices
                    2D matrix of four columns [IL, XL, Z, Value]

    Note:
        Negative z is used in the vertical direction
    """

    if np.ndim(seis3dmat) != 3:
        vis_msg.print('ERROR in retrieveSeisZSliceFrom3DMat: 3D seismic matrix expected', type='error')
        sys.exit()

    if seisinfo is None:
        vis_msg.print('WARNING in retrieveSeisZSliceFrom3DMat: Survey info auto-generated from 3D seismic matrix',
                      type='warning')
        seisinfo = createSeisInfoFrom3DMat(seis3dmat)

    if zsls is None:
        vis_msg.print('WARNING in retrieveSeisZSliceFrom3DMat: to retrieve all z slices', type='warning')
        seis2dmat = convertSeis3DMatTo2DMat(seis3dmat, seisinfo)
        seisloc = seis2dmat[:, 0:3]
        seisdata = seis2dmat[:, 3]
        seisdata = np.reshape(seisdata, [len(seisdata), 1])
        seisdata = np.concatenate((seisloc, seisdata), axis=1)
        return seisdata

    if np.ndim(zsls) != 1:
        vis_msg.print('ERROR in retrieveSeisZSliceFrom3DMat: 1D array of z slices expected', type='error')
        sys.exit()

    zstart = seisinfo['ZStart']
    zstep = seisinfo['ZStep']
    znum, xlnum, inlnum = np.shape(seis3dmat)
    if znum == 1:
        zstep = -1

    nzsls = len(zsls)
    if verbose:
        print('Retrieve ' + str(nzsls) + ' z slices')

    seisdata = np.zeros([nzsls, xlnum, inlnum], np.float32)
    for i in range(nzsls):
        #
        if verbose:
            sys.stdout.write('\r>>> Process %.1f%%' % (float(i+1) / float(nzsls) * 100.0))
            sys.stdout.flush()
        #
        z = zsls[i]
        idx = np.round((z - zstart) / zstep).astype(np.int32)
        if idx >= 0 and idx < znum:
            seisdata[i, :, :] = seis3dmat[idx, :, :]
    #
    if verbose:
        sys.stdout.write(' Done\n')

    all_sample = np.arange(0, inlnum * xlnum * nzsls)
    seisinl = seisinfo['ILRange']
    seisinl = np.reshape(seisinl[(all_sample / xlnum / nzsls).astype(int)], [-1, 1])
    seisxl = seisinfo['XLRange']
    seisxl = np.reshape(seisxl[((all_sample / nzsls) % xlnum).astype(int)], [-1, 1])
    seisz = zsls
    seisz = np.reshape(seisz[(all_sample % nzsls).astype(int)], [-1, 1])
    #
    seisdata = seisdata.transpose()
    seisdata = np.reshape(seisdata, [1, nzsls * xlnum * inlnum])
    seisdata = seisdata.transpose()

    seisdata = np.concatenate((seisinl, seisxl, seisz, seisdata), axis=1)

    return seisdata


def retrieveSeisWindowFrom2DMat(seis2dmat, wdctrs, datacol=3,
                                inlcol=0, xlcol=1, zcol=2,
                                wdinl=0, wdxl=0, wdz=0,
                                verbose=True, qpgsdlg=None):
    """
    Retrieve seismic data in a specified window from 2D matrix (Seis2DMat)

    Args:
        seis2dmat:  seismic data in 2D matrix of at least four columns [IL, XL, Z, Value, ...]
        wdctrs:     list of window centers in 2D matrix of three columns [IL, XL, Z]
        datacol:    index of data column for retrieval (indexing from 0)
                    Default is the fourth column (3)
        inlcol:     index of inline column. Default is the first column (0)
        xlcol:      index of crossline column. Default is the second column (1)
        zcol:       index of z column. Default is the third column (2)
        wdinl:      window radius along inline direction. Default is 0
        wdxl:       window radius along crossline direction. Default is 0
        wdz:        window radius along z direction. Default is 0
        verbose:    flag for message display. Default is True
        qpgsdlg:    QProgressDialog for displaying progress. Default is None

    Return:
        seisdata:   seismic data in a specified window
                    2D matrix of four columns [IL, XL, Z, WD1, WD2, WD3, ...]

    Note:
        Negative z is used in the vertical direction
    """

    # Check input matrix
    if np.ndim(seis2dmat) != 2:
        vis_msg.print('ERROR in retrieveSeisWindowFrom2DMat: 2D seismic matrix expected', type='error')
        sys.exit()
    if datacol < 0 or len(seis2dmat[0, :]) <= datacol:
        vis_msg.print('ERROR in retrieveSeisWindowFrom2DMat: Not data column found in 2D seismic matrix',
                      type='error')
        sys.exit()
    if inlcol < 0 or len(seis2dmat[0, :]) < inlcol:
        vis_msg.print('ERROR in retrieveSeisWindowFrom2DMat: No inline column found in 2D seismic matrix',
                      type='error')
        sys.exit()
    if xlcol < 0 or len(seis2dmat[0, :]) < xlcol:
        vis_msg.print('ERROR in retrieveSeisWindowFrom2DMat: No crossline column found in 2D seismic matrix',
                      type='error')
        sys.exit()
    if zcol < 0 or len(seis2dmat[0, :]) < zcol:
        vis_msg.print('ERROR in retrieveSeisWindowFrom2DMat: No z column found in 2D seismic matrix', type='error')
        sys.exit()

    if np.ndim(wdctrs) != 2:
        vis_msg.print('ERROR in retrieveSeisWindowFrom2DMat: 2D matrix of window centers expected', type='error')
        sys.exit()
    if len(wdctrs[0, :]) < 3:
        vis_msg.print('ERROR in retrieveSeisWindowFrom2DMat: 3 columns expected in 2D window-center matrix',
                      type='error')
        sys.exit()

    if wdinl < 0 or wdxl < 0 or wdz < 0:
        vis_msg.print('ERROR in retrieveSeisWindowFrom2DMat: Wrong window size', type='error')
        sys.exit()

    seisinfo = getSeisInfoFrom2DMat(seis2dmat,
                                    inlcol=inlcol, xlcol=xlcol, zcol=zcol)
    seis3dmat = convertSeis2DMatTo3DMat(seis2dmat,
                                        datacol=datacol,
                                        inlcol=inlcol, xlcol=xlcol, zcol=zcol)
    inlstart = seisinfo['ILStart']
    inlstep = seisinfo['ILStep']
    inlnum = seisinfo['ILNum']
    xlstart = seisinfo['XLStart']
    xlstep = seisinfo['XLStep']
    xlnum = seisinfo['XLNum']
    zstart = seisinfo['ZStart']
    zstep = seisinfo['ZStep']
    znum = seisinfo['ZNum']
    if inlnum == 1:
        inlstep = 1
    if xlnum == 1:
        xlstep = 1
    if znum == 1:
        zstep = -1

    nwdctr, _ = np.shape(wdctrs)
    if verbose:
        print('Retrieve ' + str(nwdctr) + ' window centers')

    wdinlsize = 2 * wdinl + 1
    wdxlsize = 2 * wdxl + 1
    wdzsize = 2 * wdz + 1
    wdsize = wdinlsize * wdxlsize * wdzsize

    if qpgsdlg is not None:
        qpgsdlg.setMaximum(nwdctr)

    seisdata = np.zeros([nwdctr, 3+wdsize], np.float32)
    for i in range(nwdctr):
        #
        if qpgsdlg is not None:
            QtCore.QCoreApplication.instance().processEvents()
            qpgsdlg.setValue(i)
        #
        if verbose:
            sys.stdout.write('\r>>> Process %.1f%%' % (float(i) / float(nwdctr) * 100.0))
            sys.stdout.flush()
        #
        inl = wdctrs[i, 0]
        xl = wdctrs[i, 1]
        z = wdctrs[i, 2]

        inlidx = np.round((inl - inlstart) / inlstep).astype(np.int32)
        xlidx = np.round((xl - xlstart) / xlstep).astype(np.int32)
        zidx = np.round((z - zstart) / zstep).astype(np.int32)

        if inlidx >= wdinl and inlidx < inlnum-wdinl and xlidx >= wdxl and xlidx < xlnum-wdxl \
                and zidx >= wdz and zidx < znum-wdz:
            idx = inlidx * xlnum * znum + xlidx * znum + zidx
            seisdata[i, 0] = seis2dmat[idx, inlcol]
            seisdata[i, 1] = seis2dmat[idx, xlcol]
            seisdata[i, 2] = seis2dmat[idx, zcol]
            seiswd = seis3dmat[zidx-wdz:zidx+wdz+1, xlidx-wdxl:xlidx+wdxl+1 ,inlidx-wdinl:inlidx+wdinl+1]
            seisdata[i, 3:3+wdsize] = np.reshape(seiswd, [1, wdsize])

            # for a, b, c in [(a, b, c) for a in range(wdzsize) for b in range(wdxlsize) for c in range(wdinlsize)]:
            #     seisdata[i, 3+a*wdxlsize*wdinlsize+b*wdinlsize+c] = \
            #         seis2dmat[idx+(c-wdinl)*xlnum*znum+(b-wdxl)*znum+(a-wdz), datacol]
    #
    if qpgsdlg is not None:
        qpgsdlg.setValue(nwdctr)
    if verbose:
        sys.stdout.write(' Done\n')

    return seisdata


def assignSeisWindowTo2DMat(seis2dmat, seiswd, wdinl=0, wdxl=0, wdz=0, verbose=True):
    """
    Assign seismic patches into Seis2DMat

    Args:
        seis2dmat:  Seismic data in a 2D array
        seiswd:     Seismic patches as a 2D array, with inline/crossline/z in column 1/2/3.
                    The pixel number is equal to (2*wdinl+1) * (2*wdxl+1) * (2*wdz+1)
        wdinl:      Window radius along inline. Default is 0
        wdxl:       Window radius along crossline. Default is 0
        wdz:        Window radius along z. Default is 0
        verbose:

    Return:
         Seis2DMat with new values assigned
    """

    seisinfo = getSeisInfoFrom2DMat(seis2dmat)
    seis3dmat = convertSeis2DMatTo3DMat(seis2dmat)
    #
    inlstart = seisinfo['ILStart']
    inlstep = seisinfo['ILStep']
    inlnum = seisinfo['ILNum']
    xlstart = seisinfo['XLStart']
    xlstep = seisinfo['XLStep']
    xlnum = seisinfo['XLNum']
    zstart = seisinfo['ZStart']
    zstep = seisinfo['ZStep']
    znum = seisinfo['ZNum']
    #

    #
    nwd = np.shape(seiswd)[0]
    inlsize = 2 * wdinl + 1
    xlsize = 2 * wdxl + 1
    zsize = 2 * wdz + 1
    if verbose:
        print('Assign ' + str(nwd) + ' windows')
    for i in range(nwd):
        #
        if verbose:
            sys.stdout.write('\r>>> Process %.1f%%' % (float(i) / float(nwd) * 100.0))
            sys.stdout.flush()
        #
        inl = seiswd[i, 0]
        xl = seiswd[i, 1]
        z = seiswd[i, 2]
        inlidx = int((inl - inlstart) / inlstep)
        xlidx = int((xl-xlstart) / xlstep)
        zidx = int((z-zstart) / zstep)
        #
        # seisdata = retrieveSeisWindowFrom3DMat(seis3dmat, seiswd[i:i+1, 0:3],
        #                                        seisinfo=seisinfo,
        #                                        wdinl=wdinl, wdxl=wdxl, wdz=wdz,
        #                                        verbose=False)
        # for j in range(inlsize*xlsize*zsize):
        #     if seisdata[0, 3+j] == 0.0:
        #         seisdata[0, 3+j] = seiswd[i, 3+j]
        #     seisdata[0, 3+j] = 0.5 * (seisdata[0, 3+j]+seiswd[i, 3+j])
        # seisdata = np.reshape(seisdata[0, 3:], [zsize, xlsize, inlsize])
        #
        if inlidx >= wdinl and inlidx < inlnum - wdinl and xlidx >= wdxl and xlidx < xlnum - wdxl \
                and zidx >= wdz and zidx < znum - wdz:
            seisdata = seis3dmat[zidx-wdz:zidx+wdz+1, xlidx-wdxl:xlidx+wdxl+1, inlidx-wdinl:inlidx+wdinl+1]
            seisdata = np.reshape(seisdata, [inlsize*xlsize*zsize])
            #
            for j in range(inlsize * xlsize * zsize):
                if seisdata[j] == 0.0:
                    seisdata[j] = seiswd[i, 3 + j]
                seisdata[j] = 0.5 * (seisdata[j] + seiswd[i, 3 + j])
            seisdata = np.reshape(seisdata, [zsize, xlsize, inlsize])
            #
            seis3dmat[zidx-wdz:zidx+wdz+1, xlidx-wdxl:xlidx+wdxl+1, inlidx-wdinl:inlidx+wdinl+1] = seisdata
    #
    if verbose:
        sys.stdout.write(' Done\n')
    #
    return convertSeis3DMatTo2DMat(seis3dmat, seisinfo)



def retrieveSeisWindowFrom3DMat(seis3dmat, wdseeds, seedloc='center', seisinfo=None,
                                wdinl=0, wdxl=0, wdz=0,
                                verbose=True, qpgsdlg=None):
    """
    Retrieve seismic data in a specified window from 3D matrix (Seis3DMat)

    Args:
        seis3dmat:  seismic data in 3D matrix [Z/XL/IL]
        wdseeds:    list of window centers in 2D matrix of three columns [IL, XL, Z]
        seisinfo:   basic information of 3D seismic survey
                    Auto-generated from 3D seismic matrix if not specified
        seedloc:    location of the seeds: center or top
        wdinl:      window size (from top) or radius (from center) along inline direction. Default is 0
        wdxl:       window size (from top) or radius (from center) along crossline direction. Default is 0
        wdz:        window size (from top) or radius (from center) along z direction. Default is 0
        verbose:    flag for message display. Default is True
        qpgsdlg:    QProgressDialog for displaying progress. Default is None

    Return:
        seisdata:   seismic data on specified samples
                    2D matrix of four columns [IL, XL, Z, Value]

    Note:
        Negative z is used in the vertical direction
    """

    if np.ndim(seis3dmat) != 3:
        vis_msg.print('ERROR in retrieveSeisWindowFrom3DMat: 3D seismic matrix expected', type='error')
        sys.exit()

    if seisinfo is None:
        vis_msg.print('WARNING in retrieveSeisWindowFrom3DMat: Survey info auto-generated from 3D seismic matrix',
                      type='warning')
        seisinfo = createSeisInfoFrom3DMat(seis3dmat)

    if np.ndim(wdseeds) != 2:
        vis_msg.print('ERROR in retrieveSeisWindowFrom3DMat: 2D matrix of window centers expected',
                      type='warning')
        sys.exit()
    if len(wdseeds[0, :]) < 3:
        vis_msg.print('ERROR in retrieveSeisWindowFrom3DMat: 3-column window-center list expected',
                      type='error')
        sys.exit()

    if wdinl < 0 or wdxl < 0 or wdz < 0:
        vis_msg.print('ERROR in retrieveSeisWindowFrom3DMat: Wrong window size', type='error')
        sys.exit()

    inlstart = seisinfo['ILStart']
    inlstep = seisinfo['ILStep']
    inlnum = seisinfo['ILNum']
    xlstart = seisinfo['XLStart']
    xlstep = seisinfo['XLStep']
    xlnum = seisinfo['XLNum']
    zstart = seisinfo['ZStart']
    zstep = seisinfo['ZStep']
    znum = seisinfo['ZNum']
    if inlnum == 1:
        inlstep = 1
    if xlnum == 1:
        xlstep = 1
    if znum == 1:
        zstep = -1

    nwdctr, _ = np.shape(wdseeds)
    if verbose:
        print('Retrieve ' + str(nwdctr) + ' window centers')

    wdinlsize = 0
    wdxlsize = 0
    wdzsize = 0
    if seedloc == 'center':
        wdinlsize = 2 * wdinl + 1
        wdxlsize = 2 * wdxl + 1
        wdzsize = 2 * wdz + 1
    else:
        wdinlsize = wdinl
        wdxlsize = wdxl
        wdzsize = wdz
    wdsize = wdinlsize * wdxlsize * wdzsize

    if qpgsdlg is not None:
        qpgsdlg.setMaximum(nwdctr)

    seisdata = np.zeros([nwdctr, 3 + wdsize], np.float32)
    for i in range(nwdctr):
        #
        if qpgsdlg is not None:
            QtCore.QCoreApplication.instance().processEvents()
            qpgsdlg.setValue(i)
        #
        if verbose:
            sys.stdout.write('\r>>> Process %.1f%%' % (float(i) / float(nwdctr) * 100.0))
            sys.stdout.flush()
        #
        inl = wdseeds[i, 0]
        xl = wdseeds[i, 1]
        z = wdseeds[i, 2]

        inlidx = np.round((inl - inlstart) / inlstep).astype(np.int32)
        xlidx = np.round((xl - xlstart) / xlstep).astype(np.int32)
        zidx = np.round((z - zstart) / zstep).astype(np.int32)

        if seedloc == 'center' and \
                inlidx >= wdinl and inlidx < inlnum-wdinl and \
                xlidx >= wdxl and xlidx < xlnum-wdxl and \
                zidx >= wdz and zidx < znum-wdz:
            seisdata[i, 0] = inlidx * inlstep + inlstart
            seisdata[i, 1] = xlidx * xlstep + xlstart
            seisdata[i, 2] = zidx * zstep + zstart
            seiswd = seis3dmat[zidx-wdz:zidx+wdz+1, xlidx-wdxl:xlidx+wdxl+1 ,inlidx-wdinl:inlidx+wdinl+1]
            seisdata[i, 3:3+wdsize] = np.reshape(seiswd, [1, wdsize])

        if seedloc != 'center' and \
                inlidx >= 0 and inlidx <= inlnum - wdinl and \
                xlidx >= 0 and xlidx <= xlnum - wdxl and \
                zidx >= 0 and zidx <= znum - wdz:
            seisdata[i, 0] = inlidx * inlstep + inlstart
            seisdata[i, 1] = xlidx * xlstep + xlstart
            seisdata[i, 2] = zidx * zstep + zstart
            seiswd = seis3dmat[zidx:zidx + wdz, xlidx:xlidx + wdxl,
                     inlidx:inlidx + wdinl]
            seisdata[i, 3:3 + wdsize] = np.reshape(seiswd, [1, wdsize])
    #
    #
    if qpgsdlg is not None:
        qpgsdlg.setValue(nwdctr)
    if verbose:
        sys.stdout.write(' Done\n')

    return seisdata


def retrieveSeisILTraceFrom2DMat(seis2dmat, trctrs, wdxl=0, wdz=0, verbose=False):
    """
    Retrieve inline traces from 2D matrix (Seis2DMat)

    Args:
        seis2dmat:  Seismic data in 2D matrix
        trctrs:     List of trace locations in a 2D array, [crossline, z]
        seisinfo:   Seismic information dictionary as defined above
        wdxl:      Retrieval radius along crossline direction. Default is 0
        wdz:       Retrieval radius along z direction. Default is 0
        verbose:    Verbose of message display. Default is False

    Return:
        Retrieved traces as a 2D array, with each row representing a trace
        Length = z_sample * (2*wdxl+1) * (2*wdz+1)
        Trace location is at column 1 (inline) and 2 (z)
    """
    #
    ntrace = np.shape(trctrs)[0]
    seis3dmat = convertSeis2DMatTo3DMat(seis2dmat)
    seisinfo = getSeisInfoFrom2DMat(seis2dmat)
    znum, xlnum, inlnum = np.shape(seis3dmat)
    #
    rst = np.zeros([ntrace, 2+ inlnum*(2*wdxl+1)*(2*wdz+1)])
    rst[:, 0:2] = trctrs
    #
    if verbose:
        print('Retrieve ' + str(ntrace) + ' z traces')
    #
    for i in range(ntrace):
        #
        if verbose:
            sys.stdout.write('\r>>> Process %.1f%%' % (float(i) / float(ntrace) * 100.0))
            sys.stdout.flush()
        #
        xl = trctrs[i, 0]
        z = trctrs[i, 1]
        xl = (xl - seisinfo['XLStart']) / seisinfo['XLStep']
        xl = int(xl)
        z = (z - seisinfo['ZStart']) / seisinfo['ZStep']
        z = int(z)
        if (xl-wdxl)>=0 and (xl+wdxl)<seisinfo['XLNum'] and (z-wdz)>=0 and (z+wdz)<seisinfo['ZNum']:
            dat = seis3dmat[z-wdz:z+wdz+1, xl-wdxl:xl+wdxl+1, :]
            rst[i, 2:] = np.reshape(dat, [1, -1])
    #
    if verbose:
        sys.stdout.write(' Done\n')
    #
    return rst


def retrieveSeisILTraceFrom3DMat(seis3dmat, trctrs, wdxl=0, wdz=0, seisinfo=None, verbose=False):
    """
    Retrieve inline traces from 3D matrix (Seis3DMat)

    Args:
        seis3dmat:  Seismic data in 3D matrix
        trctrs:     List of trace locations in a 2D array, [crossline, z]
        seisinfo:   Seismic information dictionary as defined above
        wdxl:      Retrieval radius along crossline direction. Default is 0
        wdz:       Retrieval radius along z direction. Default is 0
        verbose:    Verbose of message display. Default is False

    Return:
        Retrieved traces as a 2D array, with each row representing a trace
        Length = z_sample * (2*wdxl+1) * (2*wdz+1)
        Trace location is at column 1 (inline) and 2 (z)
    """
    #
    if (seisinfo is None) or (checkSeisInfo(seisinfo) is False):
        vis_msg.print("Warning in retrieveSeisILTraceFrom3DMat: SeisInfo automatically generated", type='warning')
        seisinfo = createSeisInfoFrom3DMat(seis3dmat)
    #
    ntrace = np.shape(trctrs)[0]
    znum, xlnum, inlnum = np.shape(seis3dmat)
    #
    rst = np.zeros([ntrace, 2+ inlnum*(2*wdxl+1)*(2*wdz+1)])
    rst[:, 0:2] = trctrs
    #
    if verbose:
        print('Retrieve ' + str(ntrace) + ' z traces')
    #
    for i in range(ntrace):
        #
        if verbose:
            sys.stdout.write('\r>>> Process %.1f%%' % (float(i) / float(ntrace) * 100.0))
            sys.stdout.flush()
        #
        xl = trctrs[i, 0]
        z = trctrs[i, 1]
        xl = (xl - seisinfo['XLStart']) / seisinfo['XLStep']
        xl = int(xl)
        z = (z - seisinfo['ZStart']) / seisinfo['ZStep']
        z = int(z)
        if (xl-wdxl)>=0 and (xl+wdxl)<seisinfo['XLNum'] and (z-wdz)>=0 and (z+wdz)<seisinfo['ZNum']:
            dat = seis3dmat[z-wdz:z+wdz+1, xl-wdxl:xl+wdxl+1, :]
            rst[i, 2:] = np.reshape(dat, [1, -1])
    #
    if verbose:
        sys.stdout.write(' Done\n')
    #
    return rst


def retrieveSeisXLTraceFrom2DMat(seis2dmat, trctrs, wdinl=0, wdz=0, verbose=False):
    """
    Retrieve crossline traces from 3D matrix (Seis3DMat)

    Args:
        seis2dmat:  Seismic data in 2D matrix
        trctrs:     List of trace locations in a 2D array, [inline, z]
        seisinfo:   Seismic information dictionary as defined above
        wdinl:      Retrieval radius along inline direction. Default is 0
        wdz:       Retrieval radius along z direction. Default is 0
        verbose:    Verbose of message display. Default is False

    Return:
        Retrieved traces as a 2D array, with each row representing a trace
        Length = z_sample * (2*wdinl+1) * (2*wdz+1)
        Trace location is at column 1 (inline) and 2 (z)
    """
    #
    ntrace = np.shape(trctrs)[0]
    seis3dmat = convertSeis2DMatTo3DMat(seis2dmat)
    seisinfo = getSeisInfoFrom2DMat(seis2dmat)
    znum, xlnum, inlnum = np.shape(seis3dmat)
    #
    rst = np.zeros([ntrace, 2+ xlnum*(2*wdinl+1)*(2*wdz+1)])
    rst[:, 0:2] = trctrs
    #
    if verbose:
        print('Retrieve ' + str(ntrace) + ' z traces')
    #
    for i in range(ntrace):
        #
        if verbose:
            sys.stdout.write('\r>>> Process %.1f%%' % (float(i) / float(ntrace) * 100.0))
            sys.stdout.flush()
        #
        inl = trctrs[i, 0]
        z = trctrs[i, 1]
        inl = (inl - seisinfo['ILStart']) / seisinfo['ILStep']
        inl = int(inl)
        z = (z - seisinfo['ZStart']) / seisinfo['ZStep']
        z = int(z)
        if (inl-wdinl)>=0 and (inl+wdinl)<seisinfo['ILNum'] and (z-wdz)>=0 and (z+wdz)<seisinfo['ZNum']:
            dat = seis3dmat[z-wdz:z+wdz+1, :, inl-wdinl:inl+wdinl+1]
            rst[i, 2:] = np.reshape(dat, [1, -1])
    #
    if verbose:
        sys.stdout.write(' Done\n')
    #
    return rst


def retrieveSeisXLTraceFrom3DMat(seis3dmat, trctrs, wdinl=0, wdz=0, seisinfo=None, verbose=False):
    """
    Retrieve crossline traces from 3D matrix (Seis3DMat)

    Args:
        seis3dmat:  Seismic data in 3D matrix
        trctrs:     List of trace locations in a 2D array, [inline, z]
        seisinfo:   Seismic information dictionary as defined above
        wdinl:      Retrieval radius along inline direction. Default is 0
        wdz:       Retrieval radius along z direction. Default is 0
        verbose:    Verbose of message display. Default is False

    Return:
        Retrieved traces as a 2D array, with each row representing a trace
        Length = z_sample * (2*wdinl+1) * (2*wdz+1)
        Trace location is at column 1 (inline) and 2 (z)
    """
    #
    if (seisinfo is None) or (checkSeisInfo(seisinfo) is False):
        vis_msg.print("Warning in retrieveSeisXLTraceFrom3DMat: SeisInfo automatically generated",
                      type='warning')
        seisinfo = createSeisInfoFrom3DMat(seis3dmat)
    #
    ntrace = np.shape(trctrs)[0]

    znum, xlnum, inlnum = np.shape(seis3dmat)
    #
    rst = np.zeros([ntrace, 2+ xlnum*(2*wdinl+1)*(2*wdz+1)])
    rst[:, 0:2] = trctrs
    #
    if verbose:
        print('Retrieve ' + str(ntrace) + ' z traces')
    #
    for i in range(ntrace):
        #
        if verbose:
            sys.stdout.write('\r>>> Process %.1f%%' % (float(i) / float(ntrace) * 100.0))
            sys.stdout.flush()
        #
        inl = trctrs[i, 0]
        z = trctrs[i, 1]
        inl = (inl - seisinfo['ILStart']) / seisinfo['ILStep']
        inl = int(inl)
        z = (z - seisinfo['ZStart']) / seisinfo['ZStep']
        z = int(z)
        if (inl-wdinl)>=0 and (inl+wdinl)<seisinfo['ILNum'] and (z-wdz)>=0 and (z+wdz)<seisinfo['ZNum']:
            dat = seis3dmat[z-wdz:z+wdz+1, :, inl-wdinl:inl+wdinl+1]
            rst[i, 2:] = np.reshape(dat, [1, -1])
    #
    if verbose:
        sys.stdout.write(' Done\n')
    #
    return rst


def retrieveSeisZTraceFrom2DMat(seis2dmat, trctrs, wdinl=0, wdxl=0, verbose=False):
    """
    Retrieve z traces from 2D matrix (Seis2DMat)

    Args:
        seis2dmat:  Seismic data in 2D matrix
        trctrs:     List of trace locations in a 2D array, [inline, crossline]
        wdinl:      Retrieval radius along inline direction. Default is 0
        wdxl:       Retrieval radius along crossline direction. Default is 0
        verbose:    Verbose of message display. Default is False

    Return:
        Retrieved traces as a 2D array, with each row representing a trace
        Length = z_sample * (2*wdinl+1) * (2*wdxl+1)
        Trace location is at column 1 (inline) and 2 (crossline)
    """
    #
    ntrace = np.shape(trctrs)[0]
    seis3dmat = convertSeis2DMatTo3DMat(seis2dmat)
    seisinfo = getSeisInfoFrom2DMat(seis2dmat)
    znum, xlnum, inlnum = np.shape(seis3dmat)
    #
    rst = np.zeros([ntrace, 2+ znum*(2*wdinl+1)*(2*wdxl+1)])
    rst[:, 0:2] = trctrs
    #
    if verbose:
        print('Retrieve ' + str(trctrs) + ' z traces')
    #
    for i in range(ntrace):
        #
        if verbose:
            sys.stdout.write('\r>>> Process %.1f%%' % (float(i) / float(ntrace) * 100.0))
            sys.stdout.flush()
        #
        inl = trctrs[i, 0]
        xl = trctrs[i, 1]
        inl = (inl - seisinfo['ILStart']) / seisinfo['ILStep']
        inl = int(inl)
        xl = (xl - seisinfo['XLStart']) / seisinfo['XLStep']
        xl = int(xl)
        if (inl-wdinl)>=0 and (inl+wdinl)<seisinfo['ILNum'] and (xl-wdxl)>=0 and (xl+wdxl)<seisinfo['XLNum']:
            dat = seis3dmat[:, xl-wdxl:xl+wdxl+1, inl-wdinl:inl+wdinl+1]
            rst[i, 2:] = np.reshape(dat, [1, -1])
    #
    if verbose:
        sys.stdout.write(' Done\n')
    #
    return rst


def retrieveSeisZTraceFrom3DMat(seis3dmat, trctrs, seisinfo=None, wdinl=0, wdxl=0, verbose=False):
    """
    Retrieve z traces from 3D matrix (Seis3DMat)

    Args:
        seis3dmat:  Seismic data in 3D matrix
        trctrs:     List of trace locations in a 2D array, [inline, crossline]
        seisinfo:   Seismic information dictionary as defined above
        wdinl:      Retrieval radius along inline direction. Default is 0
        wdxl:       Retrieval radius along crossline direction. Default is 0
        verbose:    Verbose of message display. Default is False

    Return:
        Retrieved traces as a 2D array, with each row representing a trace
        Length = z_sample * (2*wdinl+1) * (2*wdxl+1)
        Trace location is at column 1 (inline) and 2 (crossline)
    """

    if (seisinfo is None) or (checkSeisInfo(seisinfo) is False):
        vis_msg.print("Warning in retrieveSeisZTraceFrom3DMat: SeisInfo automatically generated",
                      type='warning')
        seisinfo = createSeisInfoFrom3DMat(seis3dmat)
    #
    ntrace = np.shape(trctrs)[0]
    znum, xlnum, inlnum = np.shape(seis3dmat)
    #
    rst = np.zeros([ntrace, 2 + znum*(2*wdinl+1)*(2*wdxl+1)])
    rst[:, 0:2] = trctrs
    #
    if verbose:
        print('Retrieve ' + str(ntrace) + ' z traces')
    #
    for i in range(ntrace):
        #
        if verbose:
            sys.stdout.write('\r>>> Process %.1f%%' % (float(i) / float(ntrace) * 100.0))
            sys.stdout.flush()
        #
        inl = trctrs[i, 0]
        xl = trctrs[i, 1]
        inl = (inl - seisinfo['ILStart']) / seisinfo['ILStep']
        inl = int(inl)
        xl = (xl - seisinfo['XLStart']) / seisinfo['XLStep']
        xl = int(xl)
        if (inl-wdinl)>=0 and (inl+wdinl)<seisinfo['ILNum'] and (xl-wdxl)>=0 and (xl+wdxl)<seisinfo['XLNum']:
            dat = seis3dmat[:, xl-wdxl:xl+wdxl+1, inl-wdinl:inl+wdinl+1]
            rst[i, 2:] = np.reshape(dat, [1, -1])
    #
    if verbose:
        sys.stdout.write(' Done\n')
    #
    return rst


def removeOutofSurveySample(pts, inlstart=0, inlend=100, xlstart=0, xlend=100, zstart=0, zend=-1000):
    """
    Remove samples out of a given survey

    Args:
        pts:        Seismic points as 2D array, with each row representing a point.
                    Trace location is represented as inline (column 1), crossline (column 2), and z (column 3)
        inlstart:   First inline No. of the survey
        inlend:     Last inline No. of the survey
        xlstart:    First crossline No. of the survey
        xlend:      Last crossline No. of the survey
        zstart:     First z No. of the survey
        zend:       Last z No. of the survey

    Return:
        samples after removal.
    """

    pts_in = []
    npts = np.shape(pts)[0]
    for i in range(npts):
        if pts[i, 0] >= inlstart and pts[i, 0] <= inlend \
            and pts[i, 1] >= xlstart and pts[i, 1] <= xlend \
            and pts[i, 2] <= zstart and pts[i, 2] >= zend:
            pts_in.append(pts[i, :])
    return np.array(pts_in)


def removeOutofSurveyXLTrace(tcs, inlstart=0, inlend=100, zstart=0, zend=-1000):
    """
    Remove crossline traces out of a given survey

    Args:
        tcs:        Seismic traces as 2D array, with each row representing a trace.
                    Trace location is represented as inline (column 1) and crossline (column 2)
        inlstart:   First inline No. of the survey
        inlend:     Last inline No. of the survey
        zstart:    First z No. of the survey
        zend:      Last z No. of the survey

    Return:
        traces after removal.
    """

    tcs_in = []
    ntc = np.shape(tcs)[0]
    for i in range(ntc):
        if tcs[i, 0] >= inlstart and tcs[i, 0] <= inlend \
            and tcs[i, 1] <= zstart and tcs[i, 1] >= zend:
            tcs_in.append(tcs[i, :])
    return np.array(tcs_in)


def removeOutofSurveyZTrace(tcs, inlstart=0, inlend=100, xlstart=0, xlend=100):
    """
    Remove z traces out of a given survey

    Args:
        tcs:        Seismic traces as 2D array, with each row representing a trace.
                    Trace location is represented as inline (column 1) and crossline (column 2)
        inlstart:   First inline No. of the survey
        inlend:     Last inline No. of the survey
        xlstart:    First crossline No. of the survey
        xlend:      Last crossline No. of the survey

    Return:
        z traces after removal.
    """

    tcs_in = []
    ntc = np.shape(tcs)[0]
    for i in range(ntc):
        if tcs[i, 0] >= inlstart and tcs[i, 0] <= inlend \
            and tcs[i, 1] >= xlstart and tcs[i, 1] <= xlend :
            tcs_in.append(tcs[i, :])
    return np.array(tcs_in)


def isOutOfSurvey(ixzlist, inlstart=0, inlend=100, xlstart=0, xlend=100, zstart=0, zend=-1000):
    """
    check if given sample out of seismic survey

    Args:
        ixzlist:    2D array of sample lists of 3 columns [IL, XL, Z]
        inlstart:   start inline number. Default is 0
        inlend:     end inline number. Default is 100
        xlstart:    start crossline number. Default is 0
        xlend:      end crossline number. Default is 100
        zstart:     start z in ms. Default is 0
        zend:       end z in ms. Default is -1000

    Return:
        1D array for the samples. positive for out of survey; 0 for on the boundary; negative for inside

    Note:
        Negative z is used in the vertical direction
    """

    if np.ndim(ixzlist) < 2 or np.shape(ixzlist)[1] < 3:
        vis_msg.print('ERROR in isOutOfSurvey: 2-D matrix of at least 3 columns expected', type='error')
        sys.exit()

    nsample = np.shape(ixzlist)[0]
    flags = np.zeros([nsample, 1])
    for i in range(nsample):
        il = ixzlist[i, 0]
        xl = ixzlist[i, 1]
        z = ixzlist[i, 2]
        # Out of survey
        if il < inlstart or il > inlend \
            or xl < xlstart or xl > xlend \
            or z > zstart or z < zend:
            flags[i] = 1
        # Inside survey
        if il > inlstart and il < inlend \
                and xl > xlstart and xl < xlend \
                and z < zstart and z > zend:
            flags[i] = -1
    #
    return flags


def convertIXZToIJK(ixz, inlstart=0, inlstep=1, xlstart=0, xlstep=1, zstart=0, zstep=-1):
    """
    Convert inline/crossline/z [IL, XL, Z] to index [I, J, K]

    Args:
        ixz:        2D matrix of inline/crossline/z
        inlstart:   start inline
        inlstep:    inline step
        xlstart:    start crossline
        xlstep:     crossline step
        zstart:     start z in ms
        zstep:      z step in ms

    Return:
        2D array for index [I, J, K]

    Note:
        Negative z is used in the vertical direction
    """

    if np.ndim(ixz) < 2 or np.shape(ixz)[1] < 3:
        vis_msg.print('ERROR in convertIXZToIJK: 2-D matrix of at least 3 columns expected', type='error')
        sys.exit()
    if inlstep <= 0 or xlstep <= 0 or zstep >= 0:
        vis_msg.print('ERROR in convertIXZToIJK: wrong inline/crossline/z step', type='error')
        sys.exit()

    ijk = ixz
    if np.shape(ijk)[0] > 0:
        ijk[:, 0] = (ixz[:, 0] - inlstart) / inlstep
        ijk[:, 1] = (ixz[:, 1] - xlstart) / xlstep
        ijk[:, 2] = (ixz[:, 2] - zstart) / zstep

    # ijk = ijk.astype(int)

    return ijk


def convertIJKToIXZ(ijk, inlstart=0, inlstep=1, xlstart=0, xlstep=1, zstart=0, zstep=-1):
    """
    Convert index [I, J, K] to inline/crossline/z [IL, XL, Z]

    Args:
        ijk:        2D matrix of [I, J, K]
        inlstart:   start inline
        inlstep:    inline step
        xlstart:    start crossline
        xlstep:     crossline step
        zstart:     start z in ms
        zstep:      z step in ms

    Return:
        2D array for [IL, XL, Z]

    Note:
        Negative z is used in the vertical direction
    """

    if np.ndim(ijk) < 2 or np.shape(ijk)[1] < 3:
        vis_msg.print('ERROR in convertIJKToIXZ: 2-D matrix of at least 3 columns expected', type='error')
        sys.exit()

    ixz = ijk
    if np.shape(ijk)[0] > 0:
        ixz[:, 0] = ijk[:, 0] * inlstep + inlstart
        ixz[:, 1] = ijk[:, 1] * xlstep + xlstart
        ixz[:, 2] = ijk[:, 2] * zstep + zstart

    # ijk = ijk.astype(int)

    return ixz


def getRandomSample(inl, xl, z, nsample=1000, unique=False):
    """
    Get samples randomly from the given inlines, xlines, and zs

    Args:
        inl:        a list of inline sections
        xl:         a list of crossline section
        z:          a list of z sections
        nsample:    number of samples expected
        unique:     Unique samples or not. If so, No. of returned samples may be shorter than nsample

    Return:
         2D matrix with each row representing a sample
    """

    inlnum = len(inl)
    xlnum = len(xl)
    znum = len(z)
    #
    seeds = np.zeros([nsample, 3])
    #
    # get seed
    seed_idx = np.random.randint(znum * xlnum * inlnum, size=nsample)
    if unique is True:
        seed_idx = np.unique(seed_idx)
    #
    # add inline
    seeds[:, 0] = inl[(seed_idx / xlnum / znum).astype(int)]
    # add crossline
    seeds[:, 1] = xl[((seed_idx / znum) % xlnum).astype(int)]
    # add z
    seeds[:, 2] = z[(seed_idx % znum).astype(int)]
    #
    return seeds


class analysis:
    # pack all functions as a class
    #
    getSeisInfoFrom2DMat = getSeisInfoFrom2DMat
    #
    checkSeisInfo = checkSeisInfo
    makeSeisInfo = makeSeisInfo
    #
    isSeis2DMatConsistentWithSeisInfo = isSeis2DMatConsistentWithSeisInfo
    isSeis3DMatConsistentWithSeisInfo = isSeis3DMatConsistentWithSeisInfo
    #
    createSeisInfoFrom3DMat = createSeisInfoFrom3DMat
    convertSeis2DMatTo3DMat = convertSeis2DMatTo3DMat
    convertSeis3DMatTo2DMat = convertSeis3DMatTo2DMat
    convertSeisInfoTo2DMat = convertSeisInfoTo2DMat
    #
    retrieveSeisSampleFrom2DMat = retrieveSeisSampleFrom2DMat
    retrieveSeisSampleFrom3DMat = retrieveSeisSampleFrom3DMat
    retrieveSeisILSliceFrom2DMat = retrieveSeisILSliceFrom2DMat
    retrieveSeisILSliceFrom3DMat = retrieveSeisILSliceFrom3DMat
    retrieveSeisXLSliceFrom2DMat = retrieveSeisXLSliceFrom2DMat
    retrieveSeisXLSliceFrom3DMat = retrieveSeisXLSliceFrom3DMat
    retrieveSeisZSliceFrom2DMat = retrieveSeisZSliceFrom2DMat
    retrieveSeisZSliceFrom3DMat = retrieveSeisZSliceFrom3DMat
    retrieveSeisWindowFrom2DMat = retrieveSeisWindowFrom2DMat
    retrieveSeisWindowFrom3DMat = retrieveSeisWindowFrom3DMat
    retrieveSeisILTraceFrom2DMat = retrieveSeisILTraceFrom2DMat
    retrieveSeisILTraceFrom3DMat = retrieveSeisILTraceFrom3DMat
    retrieveSeisXLTraceFrom2DMat = retrieveSeisXLTraceFrom2DMat
    retrieveSeisXLTraceFrom3DMat = retrieveSeisXLTraceFrom3DMat
    retrieveSeisZTraceFrom2DMat = retrieveSeisZTraceFrom2DMat
    retrieveSeisZTraceFrom3DMat = retrieveSeisZTraceFrom3DMat
    cutSeisILSliceFrom2DMat = cutSeisILSliceFrom2DMat
    cutSeisXLSliceFrom2DMat = cutSeisXLSliceFrom2DMat
    assignSeisWindowTo2DMat = assignSeisWindowTo2DMat
    #
    removeOutofSurveySample = removeOutofSurveySample
    removeOutofSurveyXLTrace = removeOutofSurveyXLTrace
    removeOutofSurveyZTrace = removeOutofSurveyZTrace
    isOutOfSurvey = isOutOfSurvey
    convertIXZToIJK = convertIXZToIJK
    convertIJKToIXZ = convertIJKToIXZ
    #
    getRandomSample = getRandomSample




#############################################################################################
#                                                                                           #
# Author:   GeoPy Team                                                                      #
# Email:    geopy.info@gmail.com                                                            #
# Date:     April 2019                                                                      #
#                                                                                           #
#############################################################################################

# Create a window for plot seismic spectrum


from PyQt5 import QtCore, QtGui, QtWidgets
import numpy as np
import sys, os
#
sys.path.append(os.path.dirname(__file__)[:-4][:-4][:-13])
from cognitivegeo.src.basic.data import data as basic_data
from cognitivegeo.src.core.settings import settings as core_set
from cognitivegeo.src.seismic.analysis import analysis as seis_ays
from cognitivegeo.src.seismic.attribute import attribute as seis_attrib
from cognitivegeo.src.pointset.visualization import visualization as point_vis
from cognitivegeo.src.gui.configlineplotting import configlineplotting as gui_configlineplotting
from cognitivegeo.src.vis.messager import messager as vis_msg

QtWidgets.QApplication.setAttribute(QtCore.Qt.AA_EnableHighDpiScaling, True)
QtWidgets.QApplication.setAttribute(QtCore.Qt.AA_UseHighDpiPixmaps, True)


class plotvis1dseisfreq(object):

    survinfo = {}
    seisdata = {}
    linestyle = core_set.Visual['Line']
    fontstyle = core_set.Visual['Font']
    #
    iconpath = os.path.dirname(__file__)
    dialog = None
    #
    lineplottingconfig = {}


    def setupGUI(self, PlotVis1DSeisFreq):
        PlotVis1DSeisFreq.setObjectName("PlotVis1DSeisFreq")
        PlotVis1DSeisFreq.setFixedSize(500, 320)
        icon = QtGui.QIcon()
        icon.addPixmap(QtGui.QPixmap(os.path.join(self.iconpath, "icons/plotcurve.png")),
                       QtGui.QIcon.Normal,
                       QtGui.QIcon.Off)
        PlotVis1DSeisFreq.setWindowIcon(icon)
        #
        self.lwgattrib = QtWidgets.QListWidget(PlotVis1DSeisFreq)
        self.lwgattrib.setObjectName("lwgattrib")
        self.lwgattrib.setGeometry(QtCore.QRect(10, 10, 230, 240))
        self.lwgattrib.setSelectionMode(QtWidgets.QAbstractItemView.ExtendedSelection)
        #
        self.lblrange = QtWidgets.QLabel(PlotVis1DSeisFreq)
        self.lblrange.setObjectName("lblrange")
        self.lblrange.setGeometry(QtCore.QRect(260, 10, 150, 30))
        self.ldtmin = QtWidgets.QLineEdit(PlotVis1DSeisFreq)
        self.ldtmin.setObjectName("ldtmin")
        self.ldtmin.setGeometry(QtCore.QRect(260, 40, 90, 30))
        self.ldtmin.setAlignment(QtCore.Qt.AlignCenter)
        self.ldtmax = QtWidgets.QLineEdit(PlotVis1DSeisFreq)
        self.ldtmax.setObjectName("ldtmax")
        self.ldtmax.setGeometry(QtCore.QRect(400, 40, 90, 30))
        self.ldtmax.setAlignment(QtCore.Qt.AlignCenter)
        self.lblrangeto = QtWidgets.QLabel(PlotVis1DSeisFreq)
        self.lblrangeto.setObjectName("lblrangeto")
        self.lblrangeto.setGeometry(QtCore.QRect(350, 40, 50, 30))
        self.lblrangeto.setAlignment(QtCore.Qt.AlignCenter)
        #
        self.cbxlegend = QtWidgets.QCheckBox(PlotVis1DSeisFreq)
        self.cbxlegend.setObjectName("cbxlegend")
        self.cbxlegend.setGeometry(QtCore.QRect(430, 220, 60, 30))
        #
        self.btnconfigline = QtWidgets.QPushButton(PlotVis1DSeisFreq)
        self.btnconfigline.setObjectName("btnconfigline")
        self.btnconfigline.setGeometry(QtCore.QRect(260, 220, 160, 30))
        icon = QtGui.QIcon()
        icon.addPixmap(QtGui.QPixmap(os.path.join(self.iconpath, "icons/settings.png")),
                       QtGui.QIcon.Normal,
                       QtGui.QIcon.Off)
        self.btnconfigline.setIcon(icon)
        #
        self.btnplot = QtWidgets.QPushButton(PlotVis1DSeisFreq)
        self.btnplot.setObjectName("btnplot")
        self.btnplot.setGeometry(QtCore.QRect(170, 270, 160, 30))
        icon = QtGui.QIcon()
        icon.addPixmap(QtGui.QPixmap(os.path.join(self.iconpath, "icons/plotcurve.png")),
                       QtGui.QIcon.Normal,
                       QtGui.QIcon.Off)
        self.btnplot.setIcon(icon)
        #
        self.msgbox = QtWidgets.QMessageBox(PlotVis1DSeisFreq)
        self.msgbox.setObjectName("msgbox")
        _center_x = PlotVis1DSeisFreq.geometry().center().x()
        _center_y = PlotVis1DSeisFreq.geometry().center().y()
        self.msgbox.setGeometry(QtCore.QRect(_center_x - 150, _center_y - 50, 300, 100))
        #
        self.retranslateGUI(PlotVis1DSeisFreq)
        QtCore.QMetaObject.connectSlotsByName(PlotVis1DSeisFreq)



    def retranslateGUI(self, PlotVis1DSeisFreq):
        self.dialog = PlotVis1DSeisFreq
        #
        _translate = QtCore.QCoreApplication.translate
        PlotVis1DSeisFreq.setWindowTitle(_translate("PlotVis1DSeisFreq", "1D Window: Seismic Spectrum"))
        if self.checkSurvInfo() is True:
            _firstattrib = None
            for i in sorted(self.seisdata.keys()):
                if self.checkSeisData(i):
                    item = QtWidgets.QListWidgetItem(self.lwgattrib)
                    item.setText(_translate("PlotVis1DSeisFreq", i))
                    self.lwgattrib.addItem(item)
                    if _firstattrib is None:
                        _firstattrib = item
            self.lwgattrib.setCurrentItem(_firstattrib)
        self.lwgattrib.itemSelectionChanged.connect(self.changeLwgAttrib)
        #
        self.lblrange.setText(_translate("PlotVis1DSeisFreq", "Frequency range:"))
        self.lblrangeto.setText(_translate("PlotVis1DSeisFreq", "~~~"))
        if (self.checkSurvInfo() is True):
            _min = 0.0
            _max = abs(500.0 / float(self.survinfo['ZStep']))
            self.ldtmin.setText(_translate("PlotVis1DSeisFreq", str(_min)))
            self.ldtmax.setText(_translate("PlotVis1DSeisFreq", str(_max)))
        #
        self.cbxlegend.setText(_translate("Plot2DSeisXl", "Legend"))
        self.cbxlegend.setChecked(True)
        #
        self.btnconfigline.setText(_translate("PlotVis1DSeisFreq", "Spectrum Configuration"))
        self.btnconfigline.clicked.connect(self.clickBtnConfigLine)
        if (self.checkSurvInfo() is True) \
                and (self.lwgattrib.currentItem() is not None) \
                and (self.checkSeisData(self.lwgattrib.currentItem().text()) is True):
            _config = self.linestyle
            self.lineplottingconfig[self.lwgattrib.currentItem().text()] = _config
        #
        self.btnplot.setText(_translate("PlotVis1DSeisFreq", "Seismic Spectrum Viewer"))
        self.btnplot.setDefault(True)
        self.btnplot.clicked.connect(self.clickBtnPlot)


    def changeLwgAttrib(self):
        if len(self.lwgattrib.selectedItems()) > 0:
            _config = {}
            for _attrib in self.lwgattrib.selectedItems():
                if _attrib.text() in self.lineplottingconfig.keys():
                    _config[_attrib.text()] = self.lineplottingconfig[_attrib.text()]
                else:
                    _config[_attrib.text()] = self.linestyle
            self.lineplottingconfig = _config
        else:
            self.lineplottingconfig = {}


    def clickBtnConfigLine(self):
        _config = QtWidgets.QDialog()
        _gui = gui_configlineplotting()
        _gui.lineplottingconfig = self.lineplottingconfig
        _gui.setupGUI(_config)
        _config.exec()
        self.lineplottingconfig = _gui.lineplottingconfig
        _config.show()


    def clickBtnPlot(self):
        self.refreshMsgBox()
        #
        _attriblist = self.lwgattrib.selectedItems()
        if len(_attriblist) < 1:
            vis_msg.print("ERROR in PlotVis1DSeisFreq: No property selected for plot", type='error')
            QtWidgets.QMessageBox.critical(self.msgbox,
                                           '1D Window: Seismic Spectrum',
                                           'No property selected for plot')
            return
        #
        _freqmin = basic_data.str2float(self.ldtmin.text())
        _freqmax = basic_data.str2float(self.ldtmax.text())
        if _freqmin is False or _freqmax is False:
            vis_msg.print("ERROR in PlotVis1DSeisFreq: Non-float range specified for plot", type='error')
            QtWidgets.QMessageBox.critical(self.msgbox,
                                           '1D Window: Seismic Spectrum',
                                           'Non-float range specified for plot')
            return
        #
        _legendon = self.cbxlegend.isChecked()
        #
        _freqdict = {}
        _markerstylelist = []
        _markersizelist = []
        _colorlist = []
        _linestylelist = []
        _linewidthlist = []
        _specmin = 1e+10
        _specmax = 0
        for i in range(len(_attriblist)):
            print("PlotVis1DSeisFreq: Plot %d of %d property spectrum: %s"
                  % (i+1, len(_attriblist), _attriblist[i].text()))
            _data = self.seisdata[_attriblist[i].text()]
            #
            _freq = seis_attrib.calcPowerSpectrum(_data, self.survinfo['ZStep'])
            _freqdict[_attriblist[i].text()] = {}
            _freqdict[_attriblist[i].text()]['Frequency'] = _freq[:, 0]
            _freqdict[_attriblist[i].text()]['Power spectrum'] = _freq[:, 1]
            if np.min(_freq[:, 1]) < _specmin:
                _specmin = np.min(_freq[:, 1])
            if np.max(_freq[:, 1]) > _specmax:
                _specmax = np.max(_freq[:, 1])
            #
            _colorlist.append(self.lineplottingconfig[_attriblist[i].text()]['Color'].lower())
            _linewidthlist.append(self.lineplottingconfig[_attriblist[i].text()]['Width'])
            _linestylelist.append(self.lineplottingconfig[_attriblist[i].text()]['Style'].lower())
            _markerstylelist.append(self.lineplottingconfig[_attriblist[i].text()]['MarkerStyle'])
            _markersizelist.append(self.lineplottingconfig[_attriblist[i].text()]['MarkerSize'])
        #
        point_vis.crossplot2D(_freqdict,
                              colorlist=_colorlist, linestylelist=_linestylelist,
                              linewidthlist=_linewidthlist,
                              markerstylelist=_markerstylelist,
                              markersizelist=_markersizelist,
                              xfeature='Frequency', yfeature='Power spectrum',
                              xlim=[_freqmin, _freqmax], ylim=[_specmin, _specmax],
                              xlabel='Frequency', ylabel='Power spectrum', legendon=_legendon,
                              fontstyle=self.fontstyle,
                              title='Spectrum',
                              qicon=QtGui.QIcon(os.path.join(self.iconpath, "icons/logo.png")),
                              qtitle='1D Window: Seismic Spectrum')
        #
        self.dialog.close()
        return


    def refreshMsgBox(self):
        _center_x = self.dialog.geometry().center().x()
        _center_y = self.dialog.geometry().center().y()
        self.msgbox.setGeometry(QtCore.QRect(_center_x - 150, _center_y - 50, 300, 100))


    def checkSurvInfo(self):
        self.refreshMsgBox()
        #
        if seis_ays.checkSeisInfo(self.survinfo) is False:
            # print("PlotVis1DSeisFreq: Survey not found")
            # QtWidgets.QMessageBox.critical(self.msgbox,
            #                                'Plot Seismic Waveform',
            #                                'Survey not found')
            return False
        return True


    def checkSeisData(self, f):
        self.refreshMsgBox()
        #
        return seis_ays.isSeis3DMatConsistentWithSeisInfo(self.seisdata[f], self.survinfo)


if __name__ == "__main__":
    app = QtWidgets.QApplication(sys.argv)
    PlotVis1DSeisFreq = QtWidgets.QWidget()
    gui = plotvis1dseisfreq()
    gui.setupGUI(PlotVis1DSeisFreq)
    PlotVis1DSeisFreq.show()
    sys.exit(app.exec_())


import os, sys
sys.path.append(os.path.dirname(__file__)[:-6][:-4][:-13])
from cognitivegeo.src.segpy.header import FormatMeta, field
from cognitivegeo.src.segpy.types import Int32, Int16


class BinaryReelHeader(metaclass=FormatMeta):

    START_OFFSET_IN_BYTES = 3201
    LENGTH_IN_BYTES = 400

    job_id_num = field(
        Int32, offset=3201, default=0, documentation=
        "Job identification number")

    line_num = field(
        Int32, offset=3205, default=0, documentation=
        "Line number. For 3-D poststack data, this will typically contain the in-line number."
    )

    reel_num = field(
        Int32, offset=3209, default=0, documentation=
        "Reel number."
    )

    data_traces_per_ensemble = field(
        Int16, offset=3213, default=0, documentation=
        "Number of data traces per ensemble. Mandatory for prestack data."
    )

    auxiliary_traces_per_ensemble = field(
        Int16, offset=3215, default=0, documentation=
        "Number of auxiliary traces per ensemble. Mandatory for prestack data."
    )

    sample_interval = field(
        Int16, offset=3217, default=0, documentation=
        "Sample interval in microseconds (μs). Mandatory for all data types."
    )

    original_field_sample_interval = field(
        Int16, offset=3219, default=0, documentation=
        "Sample interval in microseconds (μs) of original field recording."
    )

    num_samples = field(
        Int16, offset=3221, default=0, documentation=
        "Number of samples per data trace. Mandatory for all types of data. "
        "Note: The sample interval and number of samples in the Binary File Header should be for the primary set of "
        "seismic data traces in the file."
    )

    original_field_num_samples = field(
        Int16, offset=3223, default=0, documentation=
        "Number of samples per data trace for original field recording."
    )

    data_sample_format = field(
        Int16, offset=3225, default=5, documentation=
        "Data sample format code. Mandatory for all data. "
        "1 = 4-byte IBM floating-point, "
        "2 = 4-byte, two's complement integer, "
        "3 = 2-byte, two's complement integer, "
        "4 = 4-byte fixed-point with gain (obsolete), "
        "5 = 4-byte IEEE floating-point, "
        "6 = Not currently used, "
        "7 = Not currently used, "
        "8 = 1-byte, two's complement integer."
    )

    ensemble_fold = field(
        Int16, offset=3227, default=0, documentation=
        "Ensemble fold. The expected number of data traces per trace ensemble (e.g. the CMP fold). "
        "Highly recommended for all types of data."
    )

    trace_sorting = field(
        Int16, offset=3229, default=0, documentation=
        "Trace sorting code (i.e. type of ensemble) : "
        "-1 = Other (should be explained in user Extended Textual File Header stanza, "
        "0 = Unknown, "
        "1 = As recorded (no sorting), "
        "2 = CDP ensemble, "
        "3 = Single fold continuous profile 4 = Horizontally stacked, "
        "5 = Common source point, "
        "6 = Common receiver point, "
        "7 = Common offset point, "
        "8 = Common mid-point, "
        "9 = Common conversion point. "
        "Highly recommended for all types of data."
    )

    vertical_sum_code = field(
        Int16, offset=3231, default=0, documentation=
        "Vertical sum code: "
        "1 = no sum, "
        "2 = two sum, "
        "..., "
        "N=M-1 sum (M=2to32,767)."
    )

    sweep_frequency_at_start = field(
        Int16, offset=3233, default=0, documentation=
        "Sweep frequency at start (Hz)."
    )

    sweep_frequency_at_end = field(
        Int16, offset=3235, default=0, documentation=
        "Sweep frequency at end (Hz)."
    )

    sweep_length = field(
        Int16, offset=3237, default=0, documentation=
        "Sweep length in milliseconds."
    )

    sweep_type = field(
        Int16, offset=3239, default=0, documentation=
        "Sweep type: 1 = linear, 2 = parabolic, 3 = exponential 4 = other."
    )

    sweep_trace_number = field(
        Int16, offset=3241, default=0, documentation=
        "Trace number of sweep channel."
    )

    sweep_trace_taper_length_at_start = field(
        Int16, offset=3243, default=0, documentation=
        "Sweep trace taper length at start in milliseconds."
    )

    sweep_trace_taper_length_at_end = field(
        Int16, offset=3245, default=0, documentation=
        "Sweep trace taper length at end in milliseconds."
    )

    taper_type = field(
        Int16, offset=3247, default=0, documentation=
        "Taper type: 1 = linear, 2 = cos2, 3 = other"
    )

    correlated_data_traces = field(
        Int16, offset=3249, default=0, documentation=
        "Correlated data traces: "
        "1 = no, "
        "2 = yes"
    )

    binary_gain_recovered = field(
        Int16, offset=3251, default=0, documentation=
        "Binary gain recovered: "
        "1 = no, "
        "2 = yes"
    )

    amplitude_recovery_method = field(
        Int16, offset=3253, default=0, documentation=
        "Amplitude recovery method: "
        "1 = none, "
        "2 = spherical divergence, "
        "3 = AGC, "
        "4 = other"
    )

    measurement_system = field(
        Int16, offset=3255, default=0, documentation=
        "Measurement system: Highly recommended for all types of data. "
        "If Location Data stanzas are included in the file, this entry must agree with the Location Data stanza. "
        "If there is a disagreement, the last Location Data stanza is the controlling authority. "
        "1 = Meters, "
        "2 = Feet"
    )

    impulse_signal_polarity = field(
        Int16, offset=3257, default=0, documentation=
        "Impulse signal polarity : "
        "1 = Increase in pressure or upward geophone case movement gives negative number on tape, "
        "2 = Increase in pressure or upward geophone case movement gives positive number on tape. "
    )

    vibratory_polarity_code = field(
        Int16, offset=3259, default=0, documentation=
        "Vibratory polarity code: "
        "Seismic signal lags pilot signal by: "
        "1 = 337.5° to 22.5°, "
        "2 = 22.5° to 67.5°, "
        "3 = 67.5° to 112.5°, "
        "4 = 112.5° to 157.5°, "
        "5 = 157.5° to 202.5°, "
        "6 = 202.5°to 247.5°,  "
        "7 = 247.5° to 292.5°, "
        "8 = 292.5° to 337.5°.")

    format_revision_num = field(
        Int16, offset=3501, default=0x100, documentation=
        "SEG Y Format Revision Number. "
        "This is a 16-bit unsigned value with a Q- point between the first and second bytes. "
        "Thus for SEG Y Revision 1.0, as defined in this document, this will be recorded as 010016. This field is "
        "mandatory for all versions of SEG Y, although a value of zero indicates “traditional” SEG Y conforming to "
        "the 1975 standard."
    )

    fixed_length_trace_flag = field(
        Int16, offset=3503, default=0, documentation=
        "Fixed length trace flag. A value of one indicates that all traces in this SEG Y file are guaranteed to have "
        "the same sample interval and number of samples, as specified in Textual File Header bytes 3217-3218 and "
        "3221-3222. A value of zero indicates that the length of the traces in the file may vary and the number of "
        "samples in bytes 115-116 of the Trace Header must be examined to determine the actual length of each trace. "
        "This field is mandatory for all versions of SEG Y, although a value of zero indicates “traditional” SEG Y "
        "conforming to the 1975 standard."
    )

    num_extended_textual_headers = field(
        Int16, offset=3505, default=0, documentation=
        "Number of 3200-byte, Extended Textual File Header records following the Binary Header. "
        "A value of zero indicates there are no Extended Textual File Header records (i.e. this file has no Extended "
        "Textual File Header(s)). A value of -1 indicates that there are a variable number of Extended Textual File "
        "Header records and the end of the Extended Textual File Header is denoted by an ((SEG: EndText)) stanza in "
        "the final record. A positive value indicates that there are exactly that many Extended Textual File Header "
        "records. Note that, although the exact number of Extended Textual File Header records may be a useful piece "
        "of information, it will not always be known at the time the Binary Header is written and it is not mandatory "
        "that a positive value be recorded here. This field is mandatory for all versions of SEG Y, although a value "
        "of zero indicates “traditional” SEG Y conforming to the 1975 standard."
    )




#############################################################################################
#                                                                                           #
# Author:   GeoPy Team                                                                      #
# Email:    geopy.info@gmail.com                                                            #
# Date:     August 2018                                                                     #
#                                                                                           #
#############################################################################################

# Create a window for view point sets


from PyQt5 import QtCore, QtGui, QtWidgets
import os, sys
#
sys.path.append(os.path.dirname(__file__)[:-4][:-4][:-13])
from cognitivegeo.src.basic.matdict import matdict as basic_mdt
from cognitivegeo.src.core.settings import settings as core_set
from cognitivegeo.src.pointset.analysis import analysis as point_ays
from cognitivegeo.src.gui.plotvis2dpointsetcrossplt import plotvis2dpointsetcrossplt as gui_plotvis2dpointsetcrossplt

QtWidgets.QApplication.setAttribute(QtCore.Qt.AA_EnableHighDpiScaling, True)
QtWidgets.QApplication.setAttribute(QtCore.Qt.AA_UseHighDpiPixmaps, True)


class viewpointset(object):

    pointsetname = ''
    pointsetdata = {}
    linestyle = core_set.Visual['Line']
    fontstyle = core_set.Visual['Font']
    #
    iconpath = os.path.dirname(__file__)
    dialog = None


    def setupGUI(self, ViewPointSet):
        ViewPointSet.setObjectName("ViewPointSet")
        ViewPointSet.setFixedSize(510, 460)
        icon = QtGui.QIcon()
        icon.addPixmap(QtGui.QPixmap(os.path.join(self.iconpath, "icons/point.png")),
                       QtGui.QIcon.Normal,
                       QtGui.QIcon.Off)
        ViewPointSet.setWindowIcon(icon)
        #
        self.btncopy = QtWidgets.QPushButton(ViewPointSet)
        self.btncopy.setObjectName("btncopy")
        self.btncopy.setGeometry(QtCore.QRect(310, 10, 80, 30))
        self.btncopy.setEnabled(False)
        icon = QtGui.QIcon()
        icon.addPixmap(QtGui.QPixmap(os.path.join(self.iconpath, "icons/copy.png")),
                       QtGui.QIcon.Normal,
                       QtGui.QIcon.Off)
        self.btncopy.setIcon(icon)
        self.btnplot = QtWidgets.QPushButton(ViewPointSet)
        self.btnplot.setObjectName("btnplot")
        self.btnplot.setGeometry(QtCore.QRect(410, 10, 80, 30))
        self.btnplot.setEnabled(False)
        icon = QtGui.QIcon()
        icon.addPixmap(QtGui.QPixmap(os.path.join(self.iconpath, "icons/plotpoint.png")),
                       QtGui.QIcon.Normal,
                       QtGui.QIcon.Off)
        self.btnplot.setIcon(icon)
        self.twgpoint = QtWidgets.QTableWidget(ViewPointSet)
        self.twgpoint.setObjectName("twgpoint")
        self.twgpoint.setGeometry(QtCore.QRect(10, 50, 480, 380))
        self.twgpoint.setColumnCount(3)
        self.twgpoint.setSelectionMode(QtWidgets.QAbstractItemView.NoSelection)
        #
        self.msgbox = QtWidgets.QMessageBox(ViewPointSet)
        self.msgbox.setObjectName("msgbox")
        _center_x = ViewPointSet.geometry().center().x()
        _center_y = ViewPointSet.geometry().center().y()
        self.msgbox.setGeometry(QtCore.QRect(_center_x - 150, _center_y - 50, 300, 100))
        #
        self.retranslateGUI(ViewPointSet)
        QtCore.QMetaObject.connectSlotsByName(ViewPointSet)



    def retranslateGUI(self, ViewPointSet):
        self.dialog = ViewPointSet
        #
        _translate = QtCore.QCoreApplication.translate
        ViewPointSet.setWindowTitle(_translate("ViewPointSet", "View PointSet " +self.pointsetname))
        self.btncopy.setText(_translate("ViewPointSet", "Copy"))
        self.btncopy.clicked.connect(self.clickBtnCopy)
        self.btnplot.setText(_translate("ViewPointSet", "Plot"))
        self.btnplot.clicked.connect(self.clickBtnPlot)
        if self.checkPointSet():
            # self.btncopy.setEnabled(True)
            self.btnplot.setEnabled(True)
            #
            _nrow = basic_mdt.maxDictConstantRow(self.pointsetdata)
            _ncol = len(self.pointsetdata.keys())
            self.twgpoint.setRowCount(_nrow)
            self.twgpoint.setColumnCount(_ncol)
            _colheader = [key for key in self.pointsetdata.keys()]
            self.twgpoint.setHorizontalHeaderLabels(_colheader)
            # self.twgpoint.verticalHeader().hide()
            for i in range(_nrow):
                for j in range(_ncol):
                    item = QtWidgets.QTableWidgetItem()
                    item.setText(_translate("ViewPointSet", str(self.pointsetdata[_colheader[j]][i, 0])))
                    item.setFlags(QtCore.Qt.ItemIsEditable)
                    item.setTextAlignment(QtCore.Qt.AlignCenter)
                    self.twgpoint.setItem(i, j, item)


    def clickBtnCopy(self):
        self.refreshMsgBox()
        #
        _s = ""
        for i in range(self.twgpoint.rowCount()):
            for j in range(self.twgpoint.columnCount()):
                _s = _s + self.twgpoint.item(i,j).text() + '\t'
            _s = _s + '\n'
        QtGui.QGuiApplication.clipboard().setText(_s)
        if len(_s) > 0:
            QtWidgets.QMessageBox.information(self.msgbox,
                                              "View PointSet",
                                              str(len(_s)) + " points copied to clipboard")


    def clickBtnPlot(self):
        _cplt = QtWidgets.QDialog()
        _gui = gui_plotvis2dpointsetcrossplt()
        _gui.pointsetdata = {}
        _gui.pointsetdata[self.pointsetname] = self.pointsetdata
        _gui.linestyle = self.linestyle
        _gui.fontstyle = self.fontstyle
        _gui.setupGUI(_cplt)
        _cplt.exec()
        _cplt.show()


    def checkPointSet(self):
        return point_ays.checkPoint(self.pointsetdata)


    def refreshMsgBox(self):
        _center_x = self.dialog.geometry().center().x()
        _center_y = self.dialog.geometry().center().y()
        self.msgbox.setGeometry(QtCore.QRect(_center_x - 150, _center_y - 50, 300, 100))


if __name__ == "__main__":
    import sys
    app = QtWidgets.QApplication(sys.argv)
    ViewPointSet = QtWidgets.QWidget()
    gui = viewpointset()
    gui.setupGUI(ViewPointSet)
    ViewPointSet.show()
    sys.exit(app.exec_())


#############################################################################################
#                                                                                           #
# Author:   GeoPy Team                                                                      #
# Email:    geopy.info@gmail.com                                                            #
# Date:     January 2019                                                                    #
#                                                                                           #
#############################################################################################

# pointset data IO

import numpy as np
import os, sys
#
sys.path.append(os.path.dirname(__file__)[:-4][:-4][:-13])
from cognitivegeo.src.vis.messager import messager as vis_msg


__all__ = ['inputoutput']

def readPointFromAscii(asciifile, comment='#', delimiter=None,
                       inlcol=0, xlcol=1, zcol=2):
    """
    Read points from an ASCII file (by numpy.loadtxt)

    Args:
        asciifile:  An ASCII file for reading
        comment:    Comments. Default is '#'
        delimiter:  Delimiter. Default is None
        inlcol:     Index of the inline column. Default is 0
        xlcol:      Index of the crossline column. Default is 1
        zcol:       Index of the z column. Default is 2

    Return:
        2D array of the points from the ASCII file, with inline at column 0, crossline at column 1, and z at column 2
    """

    if os.path.exists(asciifile) is False:
        vis_msg.print("ERROR in readPointFromAscii: Pointset file not found", type='error')
        sys.exit()
    #
    data = np.loadtxt(asciifile, comments=comment, delimiter=delimiter)
    #
    npt, ncol = np.shape(data)
    #
    if inlcol >= ncol or inlcol < 0:
        vis_msg.print("ERROR in readPointFromAscii: Inline column index not found", type='error')
        sys.exit()
    if xlcol >= ncol or xlcol < 0:
        vis_msg.print("ERROR in readPointFromAscii: Crossline column index not found", type='error')
        sys.exit()
    if zcol >= ncol or zcol < 0:
        vis_msg.print("ERROR in readPointFromAscii: Z column index not found", type='error')
        sys.exit()
    #
    point = np.zeros([npt, ncol])
    point[:, 0:1] = data[:, inlcol:inlcol+1]
    point[:, 1:2] = data[:, xlcol:xlcol+1]
    point[:, 2:3] = data[:, zcol:zcol+1]
    # more columns
    idx = 3
    for i in range(ncol):
        if i != inlcol and i != xlcol and i != zcol:
            point[:, idx:idx+1] = data[:, i:i+1]
            idx = idx + 1
    #
    return point


class inputoutput:
    # group all functions as a single class
    readPointFromAscii = readPointFromAscii